/* 
	FILE: Box.cpp
	PURPOSE: Basic ISSIGraph hierarchy class (implementation)
	Author: Carlos Jimnez de Parga (c)
	License: GPL v3.0
*/

#include "box.h"

#include "wx/glCanvas.h"

// ----------------------------------------------------------------------------
// Object class implementation
// ----------------------------------------------------------------------------

// Main constructor

Object::Object(float tx, float ty, float tz, float width, float height)
{
	m_tx = m_x = tx;
	m_ty = m_y = ty;
	m_tz = m_z = tz;
	m_w = width;
	m_h = height;
	m_xAngle = 0.0f;
	m_yAngle = 0.0f;

	// Defaults
	
	m_MouseX = 512;
	m_MouseY = 384;
	m_UndoX = m_x;
	m_UndoY = m_y;

	m_RedoX = m_x;
	m_RedoY = m_y;

	m_Color = wxColour(255, 255, 255);
	m_Alpha = 100;

	m_isMoved = false;
	m_Toggle  = false;

}

// Standard constructor

Object::Object()
{
	m_tx = m_x = 0.0f;
	m_ty = m_y = 0.0f;
	m_tz = m_z = 1.0f;
	m_w = 0.0f;
	m_h = 0.0f;
	m_xAngle = 0.0f;
	m_yAngle = 0.0f;

	// Defaults

	m_MouseX = 512;
	m_MouseY = 384;
	m_UndoX = m_x;
	m_UndoY = m_y;

	m_RedoX = m_x;
	m_RedoY = m_y;

	m_Color = wxColour(255, 255, 255);
	m_Alpha  = 100;

	m_isMoved = false;
	m_Toggle  = false;
}

// Copy public interface

void Object::Copy(Object* obj)
{
	obj->DoCopy(m_x, m_y, m_z, m_w, m_h, m_tx, m_ty, m_tz, m_xAngle, m_yAngle, m_MouseX, m_MouseY, m_UndoX, m_UndoY, m_Color, m_Alpha, m_Name);
}

// Performs copy

void Object::DoCopy(float x, float y, float z, float w, float h, float tx, float ty, float tz, float xAngle, float yAngle,int MouseX, int MouseY, int UndoX, int UndoY, wxColour& Color, int Alpha, GLuint Name)
{
	m_x = x;
	m_y = y;
	m_z = z;
	m_w = w;
	m_h = h;
	m_tx = tx;
	m_ty = ty;
	m_tz = tz;
	m_xAngle = xAngle;
	m_yAngle = yAngle;
	m_MouseX = MouseX;
	m_MouseY = MouseY;
	m_UndoX = UndoX;
	m_UndoY = UndoY;
	m_Color = Color;
	m_Alpha = Alpha;
	m_Name = Name;
}

// Saves on XML

void Object::Save(wxString& XMLContent)
{
	
	XMLContent+= "<issiobject>"; 
	XMLContent+= _("<m_x>") + _(Tool::Math::ConvertNumber(m_x)) +  _("</m_x>");
	XMLContent+= _("<m_y>") + _(Tool::Math::ConvertNumber(m_y)) +  _("</m_y>");
	XMLContent+= _("<m_z>") + _(Tool::Math::ConvertNumber(m_z)) +  _("</m_z>");
	
	XMLContent+= _("<m_w>") + _(Tool::Math::ConvertNumber(m_w)) +  _("</m_w>");
	XMLContent+= _("<m_h>") + _(Tool::Math::ConvertNumber(m_h)) +  _("</m_h>");
	
	XMLContent+= _("<m_tx>") + _(Tool::Math::ConvertNumber(m_tx)) + _("</m_tx>");
	XMLContent+= _("<m_ty>") + _(Tool::Math::ConvertNumber(m_ty)) + _("</m_ty>");
	XMLContent+= _("<m_tz>") + _(Tool::Math::ConvertNumber(m_tz)) + _("</m_tz>");
	
	XMLContent+= _("<m_xAngle>") + _(Tool::Math::ConvertNumber(m_xAngle)) + _("</m_xAngle>");
	XMLContent+= _("<m_yAngle>") + _(Tool::Math::ConvertNumber(m_yAngle)) + _("</m_yAngle>");
	
	XMLContent+= _("<m_MouseX>") + _(Tool::Math::ConvertNumber(m_MouseX)) + _("</m_MouseX>");
	XMLContent+= _("<m_MouseY>") + _(Tool::Math::ConvertNumber(m_MouseY)) + _("</m_MouseY>");
	
	XMLContent+= _("<m_UndoX>") + _(Tool::Math::ConvertNumber(m_UndoX)) + _("</m_UndoX>");
	XMLContent+= _("<m_UndoY>") + _(Tool::Math::ConvertNumber(m_UndoY)) + _("</m_UndoY>");
	
	XMLContent+= "</issiobject>";
}

// Loads XML

void Object::Load(xml_node<> *InNode)
{
	
	m_x = atof(InNode->value());
	InNode = InNode->next_sibling();
	
	m_y = atof(InNode->value());
	InNode = InNode->next_sibling();

	m_z = atof(InNode->value());
	InNode = InNode->next_sibling();

	m_w = atof(InNode->value());
	InNode = InNode->next_sibling();

	m_h = atof(InNode->value());
	InNode = InNode->next_sibling();

	m_tx = atof(InNode->value());
	InNode = InNode->next_sibling();

	m_ty = atof(InNode->value());
	InNode = InNode->next_sibling();

	m_tz = atof(InNode->value());
	InNode = InNode->next_sibling();

	m_xAngle = atof(InNode->value());
	InNode = InNode->next_sibling();

	m_yAngle = atof(InNode->value());
	InNode = InNode->next_sibling();

	m_MouseX = atof(InNode->value());
	InNode = InNode->next_sibling();
	
	m_MouseY = atof(InNode->value());
	InNode = InNode->next_sibling();

	m_UndoX = atof(InNode->value());
	InNode = InNode->next_sibling();

	m_UndoY = atof(InNode->value());
	InNode = InNode->next_sibling();

}

// Returns width

float Object::GetWidth()
{
	return m_w;
}

// Returns height

float Object::GetHeight()
{
	return m_h;
}

// Returns X angle

float Object::GetAngleX()
{
	return m_xAngle;
}

// Returns Y angle

float Object::GetAngleY()
{
	return m_yAngle;
}

// Sets X angle

void Object::SetAngleX(float xAngle)
{
	m_xAngle = xAngle;
}

// Sets y Angle

void Object::SetAngleY(float yAngle)
{
	m_yAngle = yAngle;
}

// Gets 3D coordinates

Tool::Vector3D Object::GetCoords()
{
	return Tool::Vector3D(m_x, m_y, m_z);
}

// Sets mouse coordinates

void Object::SetMouse(int x, int y)
{
	m_MouseX = x;
	m_MouseY = y;
}

// Sets color

void Object::SetColor(wxColour& Color)
{
	m_Color = Color;
}

// Sets OpenGL alpha channel

void Object::SetAlpha(int Range)
{
	m_Alpha = Range;
}

// Gets Mouse coordinates

Tool::COORDS Object::GetMouse()
{
	Tool::COORDS Cords;
	Cords.m_x = m_MouseX;
	Cords.m_y = m_MouseY;
	return Cords;
}

// Set UNDO last movement

void Object::SetUndo(Tool::COORDS Cords)
{
	m_UndoX = Cords.m_x;
	m_UndoY = Cords.m_y;
}

// Gets UNDO last movement

Tool::COORDS Object::GetUndo()
{
	Tool::COORDS Cords;
	Cords.m_x = m_UndoX;
	Cords.m_y = m_UndoY;
	return Cords;
}

// Set REDO last movement

void Object::SetRedo(Tool::COORDS Cords)
{
	m_RedoX = Cords.m_x;
	m_RedoY = Cords.m_y;
}

// Gets REDO last movement

Tool::COORDS Object::GetRedo()
{
	Tool::COORDS Cords;
	Cords.m_x = m_RedoX;
	Cords.m_y = m_RedoY;
	return Cords;
}

bool Object::isUndoToggle()
{
	return m_Toggle;
}

void Object::SetUndoToggle(bool Toggle)
{
	m_Toggle = Toggle;
}

// Sets name or 3D openGL object's ID

void Object::SetName(GLuint Name)
{
	m_Name = Name;
}

// Returns ID or name

GLuint Object::GetName()
{
	return m_Name;
}

// ----------------------------------------------------------------------------
// Box implementation
// ----------------------------------------------------------------------------


// Default constructor

Box::Box()
{}

// Virtual destructor

Box::~Box()
{}


// Main constructor

Box::Box(float ptx, float pty, float ptz, float width, float height):Object(ptx, pty, ptz, width, height)
{
	m_Selected = false;
	m_PassR = true;
}

// Copy public interface

void Box::Copy(Box* box)
{
	Object::Copy(box);
	box->DoCopy(m_Selected);
}

// Performs copy

void Box::DoCopy(bool selected)
{
	m_Selected = selected;
}

// Compares two names

bool Box::CompareName(GLuint name)
{
	return (m_Name == name);
}

// Sets if selected

void Box::SetSelected(bool selected)
{
	m_Selected = selected;
}

// Gets if selected

bool Box::GetSelected()
{
	return m_Selected;
}

void Box::SetPassR(bool Pass)
{
	m_PassR = Pass;
}

// Restore rotation coordinates after translation
// Remarks: it gives a m_x, m_y, m_z coordinates
// according to m_tx, m_ty and m_tz
// i.e. reverse rotation transform

void Box::RestoreR(float xAngle, float yAngle)
{
	const float piXAngle = PI*xAngle/180.0f;
	const float piYAngle = PI*yAngle/180.0f;
	
	
	float AuxX, AuxY, AuxZ;
	

	// Inverse transform (Y axis)

	AuxX = m_tx*cos(piYAngle) - m_tz*sin(piYAngle);
	
	if (piYAngle!=0)
		AuxZ = (m_tx - AuxX*cos(piYAngle))/sin(piYAngle);
	else 
		AuxZ= m_tz;
		
	AuxY = m_ty;
	
	m_x = AuxX;

	// Inverse transform (X axis)

	if (piXAngle!=0)
		m_z = AuxZ*cos(piXAngle)-AuxY*sin(piXAngle);
	else m_z = AuxZ;
	
	m_y = (AuxY + m_z*sin(piXAngle))/cos(piXAngle);

	
}


// General rotate

void Box::RotateScene(float xAngle, float yAngle)
{
		
		
	const float piXAngle = PI*xAngle/180.0f;
	const float piYAngle = PI*yAngle/180.0f;
	

	// If surface moved, o control point moved then m_Pass = true
	if (m_PassR)
	{
		RestoreR(xAngle, yAngle);
		m_PassR = false;
	}
	
	
	// Rotates X axis
	
	m_ty = m_y*cos(piXAngle) - m_z*sin(piXAngle); 
	m_tz = m_y*sin(piXAngle) + m_z*cos(piXAngle); 
	m_tx = m_x;

	float AuxZ = m_tz;
	float AuxX = m_tx;
	
	// Rotates Y axis

	m_tz = AuxZ*cos(piYAngle) - AuxX*sin(piYAngle);
	m_tx = AuxZ*sin(piYAngle) + AuxX*cos(piYAngle);

	
}


// ----------------------------------------------------------------------------
// QuadricBox implementation
// ----------------------------------------------------------------------------


// Default constructor

QuadricBox::QuadricBox():Box()
{}

// Main constructor

QuadricBox::QuadricBox(float tx, float ty, float tz, float width, float height):Box(tx, ty, tz, width, height)
{
}

// Public copy interface

void QuadricBox::Copy(QuadricBox* qb)
{
	Box::Copy(qb);
}

// Spins de quadric object given x-angle and y-angle

void QuadricBox::Spin(float xSpin, float ySpin)
{
	m_xAngle+= xSpin; 
	m_yAngle+= ySpin;

	m_xAngle %= 360;
	m_yAngle %= 360;
}

// Move the quadric within the rotation

void QuadricBox::Move(int x, int y, float SceneX, float SceneY, int ScreenWidth, int ScreenHeight)
{
	
	if (m_z >= 0)
	{
		m_x = (Tool::Math::GetOGLPos(x,y, ScreenWidth, ScreenHeight).m_x + SceneX);
		m_y = (Tool::Math::GetOGLPos(x,y, ScreenWidth, ScreenHeight).m_y + SceneY);
	} else
	{
		m_x = (Tool::Math::GetOGLPos(x,y, ScreenWidth, ScreenHeight).m_x + SceneX);
		m_y = (Tool::Math::GetOGLPos(x,y, ScreenWidth, ScreenHeight).m_y + SceneY);
	}
	m_tx = m_x;
	m_ty = m_y;

	m_PassR = true;
}


// Performs paste

void QuadricBox::Paste(float x, float y, float SceneX, float SceneY)
{
	
	if (m_z >= 0)
	{
		m_x = x + SceneX;
		m_y = y + SceneY;
	} else
	{
		m_x = x + SceneX;
		m_y = y + SceneY;
	}
	
	m_tx = m_x;
	m_ty = m_y;

	m_PassR = true;
}



// Sets zooms

void QuadricBox::ZoomPlus()
{
	if (m_z < MAX_Z) 
		m_tz = ++m_z;
	
	m_PassR = true;
}

void QuadricBox::ZoomMinus()
{
	if (m_z> MIN_Z)
		m_tz = --m_z;

	m_PassR = true;
}

// Save part to XML

void QuadricBox::Save(wxString& XMLContent)
{
	const int Red = m_Color.Red();
	
	XMLContent+= "<issiobjectcolor>";
	XMLContent+= _("<m_Red>") + _(Tool::Math::ConvertNumber(Red)) + _("</m_Red>");
	
	int Green = m_Color.Green();
	XMLContent+= _("<m_Green>") + _(Tool::Math::ConvertNumber(Green)) + _("</m_Green>");
	
	int Blue = m_Color.Blue();
	XMLContent+= _("<m_Blue>") + _(Tool::Math::ConvertNumber(Blue)) + _("</m_Blue>");

	XMLContent+= _("<m_Alpha>") + _(Tool::Math::ConvertNumber(m_Alpha)) + _("</m_Alpha>");

	XMLContent+= "</issiobjectcolor>";
	Object::Save(XMLContent);
}

// Load part XML

void QuadricBox::Load(xml_node<>* InNode)
{
	int Red;
	
	Red = atof(InNode->value());
	InNode = InNode->next_sibling();

	int Green;
	Green = atof(InNode->value());
	InNode = InNode->next_sibling();

	int Blue;
	Blue = atof(InNode->value());
	InNode = InNode->next_sibling();
	
	m_Color = wxColor(Red, Green, Blue);

	int Alpha;
	Alpha = atof(InNode->value());
	
	m_Alpha = Alpha;

	Object::Load(InNode->parent()->next_sibling()->first_node());
}

// ----------------------------------------------------------------------------
// VertexBox implementation
// ----------------------------------------------------------------------------

// Default constructor

VertexBox::VertexBox():Box()
{
}

// Main constructor

VertexBox::VertexBox(float ptx, float pty, float ptz, float width, float height):Box(ptx, pty, ptz, width, height)
{
	m_PassV = true;
}

// Public copy interface

void VertexBox::Copy(VertexBox* vbox)
{
	Box::Copy(vbox);
	vbox->DoCopy(m_CorX, m_CorY);

}

// Performs copy

void VertexBox::DoCopy(bool CorX, bool CorY)
{
	m_CorX = CorX;
	m_CorY = CorY;
}

// Draws the selection box

void VertexBox::DrawBox()
{
	glPushMatrix();
	
	glEnable(GL_LINE_STIPPLE);
	glLineStipple(5,0xaaaa);
	
	glColor3f(1.0f,1.0f,0.0f);
	glTranslatef(m_tx,m_ty,m_tz);
	glBegin(GL_QUADS);
		glVertex3f(-m_w, -m_h, 0.0f);
		glVertex3f(m_w, -m_h, 0.0f);
		glVertex3f(m_w, m_h, 0.0f);
		glVertex3f(-m_w, m_h, 0.0f);
	glEnd();
		
	glDisable(GL_LINE_STIPPLE); 
	
	glPopMatrix();

}

// Moves the object according the rotation angle and sign zone

void VertexBox::Move(int x, int y, float SceneX, float SceneY, int ScreenWidth, int ScreenHeight)
{
	
	if (m_z >= 0)
	{
		m_x = Tool::Math::GetOGLPos(x,y, ScreenWidth, ScreenHeight).m_x + SceneX;
		m_y = Tool::Math::GetOGLPos(x,y, ScreenWidth, ScreenHeight).m_y + SceneY;
	} else
	{
		m_x = Tool::Math::GetOGLPos(x,y, ScreenWidth, ScreenHeight).m_x + SceneX;
		m_y = Tool::Math::GetOGLPos(x,y, ScreenWidth, ScreenHeight).m_y + SceneY;
	}
	m_tx = m_x;
	m_ty = m_y;

	m_PassR = true;
	m_PassV = true;
}

// Performs paste

void VertexBox::Paste(float x, float y,float SceneX, float SceneY)
{
	
	if (m_z >= 0)
	{
		m_x = x + SceneX;
		m_y = y + SceneY;
	} else
	{
		m_x = x + SceneX;
		m_y = y + SceneY;
	}
	m_tx = m_x;
	m_ty = m_y;

	m_PassR = true;
	m_PassV = true;
}

// Zooms the object according the rotation angle and sign zone

void VertexBox::ZoomPlus()
{
	if (m_z < MAX_Z) 
		m_tz = m_z++;

	m_PassR = true;
	m_PassV = true;	
}

void VertexBox::ZoomMinus()
{
	if (m_z > MIN_Z)
		m_tz = m_z--;

	m_PassR = true;
	m_PassV = true;
}


// Restore rotation coordinates after translation
// Remarks: it gives a m_x, m_y, m_z coordinates
// according to m_tx, m_ty and m_tz
// i.e. reverse rotation transform

void VertexBox::RestoreM()
{
	const float piXAngle = PI*m_xAngle/180.0f;
	const float piYAngle = PI*m_yAngle/180.0f;
	
	float AvgX, AvgY, AvgZ;

	// Translate to center
	
	AvgX = m_tx - m_Average.m_x;
	AvgY = m_ty - m_Average.m_y;
	AvgZ = m_tz - m_Average.m_z;
	
	float AuxX, AuxY, AuxZ;

	// Inverse transform (Y axis)

	AuxX = AvgX*cos(piYAngle) - AvgZ*sin(piYAngle);
	
	if (piYAngle!=0)
		AuxZ = (AvgX - AuxX*cos(piYAngle))/sin(piYAngle);
	else 
		AuxZ= AvgZ;
		
	AuxY = AvgY;
	
	m_x = AuxX;

	// Inverse transform (X axis)

	if (piXAngle!=0)
		m_z = AuxZ*cos(piXAngle)-AuxY*sin(piXAngle);
	else m_z = AuxZ;
	
	m_y = (AuxY + m_z*sin(piXAngle))/cos(piXAngle);

	// Translate to gravity center

	m_x+= m_Average.m_x;
	m_y+= m_Average.m_y;
	m_z+= m_Average.m_z;
	
}

// Activates when calling to RestoreM

void VertexBox::SetPassV(bool pass)
{
	m_PassV = pass;
}

// Rotates the object through X and Y axis

void VertexBox::Spin(float xSpin, float ySpin)
{
	float piXAngle, piYAngle;
		
	m_xAngle+= xSpin; 
	m_xAngle %= 360;
	piXAngle = PI*m_xAngle/180.0f;

	m_yAngle+= ySpin;
	m_yAngle %= 360;
	piYAngle = PI*m_yAngle/180.0f;

	float AvgX, AvgY, AvgZ;
	
	// If surface moved, o control point moved then m_Pass = true
	if (m_PassV)
	{
		RestoreM();
		m_PassV = false;
	}
	
	// Translate to center
	AvgX = m_x - m_Average.m_x;
	AvgY = m_y - m_Average.m_y;
	AvgZ = m_z - m_Average.m_z;
	
	// Rotates X axis
	
	m_ty = AvgY*cos(piXAngle) - AvgZ*sin(piXAngle);
	m_tz = AvgY*sin(piXAngle) + AvgZ*cos(piXAngle);
	m_tx = AvgX;

	float AuxZ = m_tz;
	float AuxX = m_tx;

	// Rotates Y axis

	m_tz = AuxZ*cos(piYAngle) - AuxX*sin(piYAngle);
	m_tx = AuxZ*sin(piYAngle) + AuxX*cos(piYAngle);
	
	m_tx+= m_Average.m_x;
	m_ty+= m_Average.m_y;
	m_tz+= m_Average.m_z;

}

//Sets average for rotation

void VertexBox::SetAverage(Tool::Vector3D& Average)
{
	m_Average = Average;
	
}

// Save part to XML

void VertexBox::Save(wxString& XMLContent)
{
	XMLContent+= "<issivertex>"; 
	XMLContent+= _("<m_CorX>") + _(Tool::Math::ConvertNumber(m_CorX)) +  _("</m_CorX>");
	XMLContent+= _("<m_CorY>") + _(Tool::Math::ConvertNumber(m_CorY)) +  _("</m_CorY>");
	XMLContent+= "</issivertex>";
		
	Object::Save(XMLContent);
}

// Loads part from XML

void VertexBox::Load(xml_node<>* Node)
{
	xml_node<>* InNode = Node->first_node();
	
	m_CorX = atoi(InNode->value());
	InNode = InNode->next_sibling();

	m_CorY = atoi(InNode->value());
	InNode = InNode->next_sibling();

}