/*****************************************************************************
   ISSIGraph: Experimental Surface Editor Prototype						 
   Copyright (C) 2013 by Carlos Jimnez de Parga								 
   
   Trabajo para el:
   MASTER EN INVESTIGACIN EN INGENIERA DE SOFTWARE Y SISTEMAS INFORMTICOS
			REPRESENTACIN GRFICA DE SUPERFICIES IMPLCITAS			
	DEPARTAMENTO DE INGENIERA DE SOFTWARE Y SISTEMAS INFORMTICOS 
							- UNED -

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
   **************************************************************************/

/* 
	FILE: Box.h
	PURPOSE: Basic ISSIGraph hierarchy class (header)
	Author: Carlos Jimnez de Parga (c)
	License: GPL v3.0
*/


#ifndef _BOX_H_
#define _BOX_H_

#include "tool.h"
#include "textout.h"
#include "../rapidxml/rapidxml.hpp"

using namespace rapidxml;
// ----------------------------------------------------------------------------
// Constants (Size of 3D Canvas)
// ----------------------------------------------------------------------------

const float MAX_X =  250.0f;
const float MIN_X = -250.0f;
const float MAX_Y =  250.0f;
const float MIN_Y = -250.0f;
const float MAX_Z =  250.0f;
const float MIN_Z = -250.0f;

const float PI = 3.141592f;

// ----------------------------------------------------------------------------
// First main base class 
// ----------------------------------------------------------------------------

class Object
{
public:
	Object(float ptx, float pty, float ptz, float width, float height);
	Object();
	void Copy(Object* obj); 
	Tool::Vector3D GetCoords();
	float GetWidth();
	float GetHeight();
	float GetAngleX(); 
	float GetAngleY();
	void SetAngleX(float xAngle);
	void SetAngleY(float yAngle);
	void SetMouse(int x, int y);
	Tool::COORDS GetMouse();
	Tool::COORDS GetUndo();
	void SetUndo(Tool::COORDS Cords);
	Tool::COORDS GetRedo();
	void SetRedo(Tool::COORDS Cords);
	GLuint GetName();
	void SetName(GLuint Name);
	void SetColor(wxColour& color);
	bool isUndoToggle();
	void SetUndoToggle(bool Toggle);
	void SetAlpha(int Range);
	virtual void Save(wxString& XMLContent);
	virtual void Load(xml_node<>* InNode);
public:
	bool m_isMoved;
protected:
	void DoCopy(float x, float y, float z, float w, float h, float tx, float ty, float tz, float xAngle, float yAngle, int MouseX, int MouseY, int UndoX, int UndoY, wxColour& Color, int Alpha, GLuint Name); 
	float m_x,m_y,m_z, m_w, m_h,m_tx,m_ty,m_tz; // The object coordinates and dimensions
	int m_xAngle,m_yAngle; // Angles of rotation around x- and y- axis
	GLuint m_Name; // The object ID
	int m_MouseX, m_MouseY;
	int m_UndoX, m_UndoY;
	int m_RedoX, m_RedoY;
	bool m_Toggle;
	wxColour m_Color;
	int m_Alpha;
};

// ----------------------------------------------------------------------------
// Second main base class
// ----------------------------------------------------------------------------


class Box : public Object
{
public:
	Box();
	virtual ~Box();
	Box(float ptx, float pty,float ptz,float width, float height);
	void Copy(Box* box); // Copy object
	virtual void DrawBox()=0; // Draws the selection box around the object
	virtual void Draw()=0; // Draws the object
	virtual void Move(int x, int y, float SceneX, float SceneY, int ScreenWidth, int ScreenHeight)=0; // Move the object
	virtual void Paste(float x, float y, float SceneX, float SceneY)=0;
	virtual void ZoomPlus()=0; // Zooms the object
	virtual void ZoomMinus()=0;
	virtual void Spin(float xSpin, float ySpin)=0; // Spin the objects
	void RestoreR(float xAngle, float yAngle); 
	void RotateScene(float xAngle, float yAngle); // General rotation
	void SetPassR(bool Pass);
	bool CompareName(GLuint Name); // To compare two IDs
	bool GetSelected(); // Activate selection
	void SetSelected(bool Selected);

protected:
	void DoCopy(bool Selected);
	bool m_Selected, m_PassR;

};


// ----------------------------------------------------------------------------
// Derived class for quadrics
// ----------------------------------------------------------------------------

class QuadricBox: public Box
{
public:
	QuadricBox();
	QuadricBox(float ptx, float pty,float ptz,float width, float height);
	void Copy(QuadricBox* Qb);
	void Move(int x, int y, float SceneX, float SceneY, int ScreenWidth, int ScreenHeight);
	void Paste(float x, float y, float SceneX, float SceneY);
	void ZoomPlus();
	void ZoomMinus();
	void Spin(float xSpin, float ySpin);
	void Save(wxString& XMLContent);
	void Load(xml_node<>* InNode);

};

// ----------------------------------------------------------------------------
// Derived class for control points and surfaces
// ----------------------------------------------------------------------------

class VertexBox : public Box
{
public:
	VertexBox();
	VertexBox(float ptx, float pty,float ptz,float width, float height);
	virtual void DrawBox();
	void Move(int x, int y, float SceneX, float SceneY, int ScreenWidth, int ScreenHeight);
	void Paste(float x, float y, float SceneX, float SceneY);
	void RestoreM();
	void ZoomPlus();
	void ZoomMinus();
	virtual void Spin(float xSpin, float ySpin);
	void SetAverage(Tool::Vector3D& Average);
	void Copy(VertexBox* Vbox);
	void Save(wxString& XMLContent);
	void Load(xml_node<>* Node);
	void SetPassV(bool pass);
protected:
	void DoCopy(bool CorX, bool CorY);
	bool  m_CorX,m_CorY, m_PassV;
	Tool::Vector3D m_Average;
};

#endif _BOX_H_