/* 
	FILE: Main.cpp
	PURPOSE: User interface (implementation)
	Remarks: This file (main.cpp) is based on the Cube.cpp wxWidget 2.9.3 OpenGL example by Julian Smart (c)
	Author: Carlos Jimnez de Parga (c)
	License: GPL v3.0 and wxWindows
*/

// For compilers that support precompilation, includes "wx.h".
#include "wx/wxprec.h"

#ifndef WX_PRECOMP
#include "wx/wx.h"
#endif
#include "main.h"

// For debug pourposes

#ifdef WIN32  
	#define _CRTDBG_MAP_ALLOC
	#include <stdlib.h>
	#include <crtdbg.h>
#endif

// Compiled built-in icons

#include "resources/appico.xpm"
#include "resources/aboutico.xpm"
#include "resources/colorico.xpm"

#include <wx/fs_inet.h>


// ----------------------------------------------------------------------------
// helper functions
// ----------------------------------------------------------------------------

static void CheckGLError()
{
    GLenum errLast = GL_NO_ERROR;

    for ( ;; )
    {
        GLenum err = glGetError();
        if ( err == GL_NO_ERROR )
            return;

        // normally the error is reset by the call to glGetError() but if
        // glGetError() itself returns an error, we risk looping forever here
        // so check that we get a different error than the last time
        if ( err == errLast )
        {
            wxLogError(wxT("OpenGL error state couldn't be reset."));
            return;
        }

        errLast = err;

        wxLogError(wxT("OpenGL error %d"), err);
    }
}

BEGIN_EVENT_TABLE(IssiLogo, wxDialog)
	EVT_TIMER(TimerLogo, IssiLogo::OnTimer)
END_EVENT_TABLE()

extern void InitXmlResource();

// IssiLogo constructor

IssiLogo::IssiLogo():wxDialog(NULL, -1, "", wxPoint(-1,-1), wxSize(420,300), wxSTAY_ON_TOP), m_Timer(this, TimerLogo)
{
	wxXmlResource::Get()->LoadObject(&m_Banner, this, wxT("Logo"), wxT("wxBannerWindow"));		
	m_Banner.SetSize(420,300);
	m_Banner.Show();
	Centre();
	m_Gauge = new wxGauge(this, -1, 2000,wxDefaultPosition, wxSize(420, 12), wxGA_HORIZONTAL);
	m_TimerValue = 0;
	m_Gauge->SetValue(m_TimerValue);
	m_Timer.Start(100);
}

// Timer loop for banner display

void IssiLogo::OnTimer(wxTimerEvent& WXUNUSED(event))
{
	m_TimerValue+=100;
	m_Gauge->SetValue(m_TimerValue);
	if (m_TimerValue >= 2000) this->Close();
}

BEGIN_EVENT_TABLE(IssiControl, wxDialog)
	EVT_BUTTON(wxID_ZOOMMINUS, OnZoomMinus)	
	EVT_BUTTON(wxID_ZOOMPLUS, OnZoomPlus)	
	EVT_BUTTON(wxID_LEFT, OnLeft)
	EVT_BUTTON(wxID_RIGHT, OnRight)	
	EVT_BUTTON(wxID_UP, OnUp)
	EVT_BUTTON(wxID_DOWN, OnDown)
	EVT_BUTTON(wxID_CENTER, OnCenter)
	EVT_BUTTON(wxID_ROTATEYPLUS, OnRotateYPlus)	
	EVT_BUTTON(wxID_ROTATEYMINUS, OnRotateYMinus)	
END_EVENT_TABLE()


// IssiControl constructor

IssiControl::IssiControl(wxFrame* Parent, IssiGLCanvas* Canvas):wxDialog(Parent, -1, _("Controller"), wxDefaultPosition, wxSize(120, 120), wxCAPTION | wxSYSTEM_MENU | wxMINIMIZE_BOX)
{
	wxSizer* Sizer = new wxGridSizer(3,3,0,0);
	SetSizer(Sizer);
	
	m_Canvas = Canvas;

	m_ButtonZoomMinus = new wxBitmapButton(this, wxID_ZOOMMINUS, wxArtProvider::GetBitmap(wxART_MINUS, wxART_BUTTON), wxDefaultPosition, wxSize(40,40));
	m_ButtonUp =  new wxBitmapButton(this, wxID_UP, wxArtProvider::GetBitmap(wxART_GO_UP, wxART_BUTTON), wxDefaultPosition, wxSize(40,40));
	m_ButtonZoomPlus =  new wxBitmapButton(this, wxID_ZOOMPLUS, wxArtProvider::GetBitmap(wxART_PLUS, wxART_BUTTON), wxDefaultPosition, wxSize(40,40));
	m_ButtonLeft =  new wxBitmapButton(this, wxID_LEFT, wxArtProvider::GetBitmap(wxART_GO_BACK, wxART_BUTTON), wxDefaultPosition, wxSize(40,40));
	m_ButtonCenter =  new wxBitmapButton(this, wxID_CENTER, wxArtProvider::GetBitmap(wxART_GO_HOME, wxART_BUTTON), wxDefaultPosition, wxSize(40,40));
	m_ButtonRight =  new wxBitmapButton(this, wxID_RIGHT, wxArtProvider::GetBitmap(wxART_GO_FORWARD, wxART_BUTTON), wxDefaultPosition, wxSize(40,40));
	m_ButtonRotateYMinus =  new wxBitmapButton(this, wxID_ROTATEYMINUS, wxArtProvider::GetBitmap(wxART_UNDO, wxART_BUTTON), wxDefaultPosition, wxSize(40,40));
	m_ButtonDown =  new wxBitmapButton(this, wxID_DOWN, wxArtProvider::GetBitmap(wxART_GO_DOWN, wxART_BUTTON), wxDefaultPosition, wxSize(40,40));
	m_ButtonRotateYPlus =  new wxBitmapButton(this, wxID_ROTATEYPLUS, wxArtProvider::GetBitmap(wxART_REDO, wxART_BUTTON), wxDefaultPosition, wxSize(40,40));
	
	Sizer->Add(m_ButtonZoomMinus, 0, wxEXPAND);
	Sizer->Add(m_ButtonUp, 0, wxEXPAND);
	Sizer->Add(m_ButtonZoomPlus, 0, wxEXPAND);
	Sizer->Add(m_ButtonLeft, 0, wxEXPAND);
	Sizer->Add(m_ButtonCenter, 0, wxEXPAND);
	Sizer->Add(m_ButtonRight, 0, wxEXPAND);
	Sizer->Add(m_ButtonRotateYMinus, 0, wxEXPAND);
	Sizer->Add(m_ButtonDown, 0, wxEXPAND);
	Sizer->Add(m_ButtonRotateYPlus, 0, wxEXPAND);


	Sizer->Fit(this);
	Sizer->SetSizeHints(this);
}

// Zooms -

void IssiControl::OnZoomMinus(wxCommandEvent& WXUNUSED(event))
{
	m_Canvas->OnSetControlOrder(wxID_ZOOMMINUS);
	m_Canvas->Refresh();
}

// Zoom +

void IssiControl::OnZoomPlus(wxCommandEvent& WXUNUSED(event))
{
	m_Canvas->OnSetControlOrder(wxID_ZOOMPLUS);
	m_Canvas->Refresh();
}

// Left

void IssiControl::OnLeft(wxCommandEvent& WXUNUSED(event))
{
	m_Canvas->OnSetControlOrder(wxID_LEFT);
	m_Canvas->Refresh();
}

// Right

void IssiControl::OnRight(wxCommandEvent& WXUNUSED(event))
{
	m_Canvas->OnSetControlOrder(wxID_RIGHT);
	m_Canvas->Refresh();
}

// Up

void IssiControl::OnUp(wxCommandEvent& WXUNUSED(event))
{
	m_Canvas->OnSetControlOrder(wxID_UP);
	m_Canvas->Refresh();
}

// Down

void IssiControl::OnDown(wxCommandEvent& WXUNUSED(event))
{
	m_Canvas->OnSetControlOrder(wxID_DOWN);
	m_Canvas->Refresh();
}

// Start-up / center

void IssiControl::OnCenter(wxCommandEvent& WXUNUSED(event))
{
	m_Canvas->OnSetControlOrder(wxID_CENTER);
	m_Canvas->Refresh();
}

// Rotate Y+

void IssiControl::OnRotateYPlus(wxCommandEvent& WXUNUSED(event))
{
	m_Canvas->OnSetControlOrder(wxID_ROTATEYPLUS);
	m_Canvas->Refresh();
}

// Rotate Y-

void IssiControl::OnRotateYMinus(wxCommandEvent& WXUNUSED(event))
{
	m_Canvas->OnSetControlOrder(wxID_ROTATEYMINUS);
	m_Canvas->Refresh();
}


// IssiAbout constructor

IssiAbout::IssiAbout(wxFrame* Parent):wxDialog(Parent, -1, _("About ISSIGraph v1.41r"), wxPoint(-1,-1),wxSize(480,400))
{
	wxStaticBox* StaticBox = new wxStaticBox(this,wxID_STATIC,_("Credits..."), wxDefaultPosition, wxSize(440,340));
	StaticBox->SetPosition(wxPoint(10,10));
	StaticBox->Show();

	wxXmlResource::Get()->LoadObject(&m_Banner, this, wxT("Logo"), wxT("wxBannerWindow"));		
	m_Banner.SetSize(420,300);
	m_Banner.SetPosition(wxPoint(20,30));
	m_Banner.Show();
	

	wxStaticText* AboutLabel = new wxStaticText(this, wxID_STATIC,_("Designed/Coded by Carlos Jimnez de Parga. This application is licensed under the GNU/GPL terms."),
		wxDefaultPosition, wxSize(230,50), wxALIGN_CENTRE);
			
	// Uses sizers

	wxBoxSizer* SizerV = new wxBoxSizer(wxVERTICAL);
	SizerV->AddSpacer(335);
	wxBoxSizer* SizerH = new wxBoxSizer(wxHORIZONTAL);
	SizerV->Add(SizerH);
	SizerH->AddSpacer(110);
	SizerH->Add(AboutLabel);
	SetSizer(SizerV);
	Centre();
}


// IssiHelp constructor


IssiHelp::IssiHelp(wxFrame* Parent):wxDialog(Parent, -1, _("ISSIGraph v1.41r HELP"), wxPoint(-1,-1),wxSize(680,500), 
	wxDEFAULT_DIALOG_STYLE | wxMAXIMIZE_BOX | wxMINIMIZE_BOX)
{

	wxHtmlWindow* Html = new wxHtmlWindow(this, wxID_ANY, wxDefaultPosition, wxSize(680, 400), wxHW_SCROLLBAR_AUTO);

	Html->SetBorders(0);
	Html->LoadPage(wxT("http://issigraph.sourceforge.net/docs.html"));
	Html->Show();	


}

// ----------------------------------------------------------------------------
// IssiGLContext
// ----------------------------------------------------------------------------

IssiGLContext::IssiGLContext(wxGLCanvas *Canvas)
             : wxGLContext(Canvas)
{
    SetCurrent(*Canvas);
	
	glewInit();

	int argc = 1;
	glutInit(&argc, NULL);

	
	// set up the parameters we want to use
	glShadeModel(GL_SMOOTH);
	glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	glEnable(GL_DEPTH_TEST);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable(GL_BLEND);
	// set viewing projection
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
	gluPerspective(60, 1.0, 1.0, 100.0);    
	glClearColor(0.5f,0.5f,0.5f,1.0f);
    CheckGLError();
	glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
	glTranslatef(0.0f, 0.0f, -50.0f);
}

// Inside draws scene loop

void IssiGLContext::DrawScene(bool Lactive, World* World)
{
	glColorMask (GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE); 
	glDepthMask (GL_TRUE);
	
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	World->Render(Lactive);   
	//glFlush();
    CheckGLError();
}


// ----------------------------------------------------------------------------
// IssiApp: the application object
// ----------------------------------------------------------------------------

IMPLEMENT_APP(IssiApp)

bool IssiApp::OnInit()
{
    // For debug pourposes
	
	#ifdef WIN32

		_CrtSetDbgFlag ( _CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF );
		_CrtSetReportMode( _CRT_ERROR, _CRTDBG_MODE_DEBUG );
	
	#endif

	if ( !wxApp::OnInit() )
        return false;

	// Internet and picture formats handlers
	
    wxFileSystem::AddHandler(new wxInternetFSHandler);
	wxImage::AddHandler(new wxJPEGHandler);
	wxImage::AddHandler(new wxPNGHandler);
	wxImage::AddHandler(new wxGIFHandler);
	
	wxXmlResource::Get()->InitAllHandlers();
	InitXmlResource();
	
	// Show logo
	
	IssiLogo LogoDlg;
	LogoDlg.ShowModal();
	
	// Start frame
		
	new IssiFrame(NULL);

    return true;
}

// Exits

int IssiApp::OnExit()
{
	delete m_glContext;

    return wxApp::OnExit();
}


// Returns context

IssiGLContext& IssiApp::GetContext(wxGLCanvas *Canvas)
{
    if ( !m_glContext )
    {
        // Create the OpenGL context for the first window which needs it:
        // subsequently created windows will all share the same context.
		m_glContext = new IssiGLContext(Canvas);
    }

    m_glContext->SetCurrent(*Canvas);

    return *m_glContext;
}

// ----------------------------------------------------------------------------
// IssiGLCanvas
// ----------------------------------------------------------------------------

BEGIN_EVENT_TABLE(IssiGLCanvas, wxGLCanvas)
	EVT_LEFT_DOWN(IssiGLCanvas::OnMouseLDown)
	EVT_RIGHT_DOWN(IssiGLCanvas::OnMouseRDown)
	EVT_MOTION(IssiGLCanvas::OnMouseMove)
	EVT_LEFT_UP(IssiGLCanvas::OnMouseLUp)
	EVT_RIGHT_UP(IssiGLCanvas::OnMouseRUp)
	EVT_MOUSEWHEEL(IssiGLCanvas::OnMouseWheel)
	EVT_PAINT(IssiGLCanvas::OnPaint)
 	EVT_KEY_DOWN(IssiGLCanvas::OnKeyDown)
	EVT_SIZE(IssiGLCanvas::OnSize)
	EVT_MENU(wxID_DELETE,IssiGLCanvas::OnPopupClick)
	EVT_MENU(wxID_TRANSFORM,IssiGLCanvas::OnPopupClick)
END_EVENT_TABLE()

IssiGLCanvas::IssiGLCanvas(wxFrame *IssiFrame)
    // With perspective OpenGL graphics, the wxFULL_REPAINT_ON_RESIZE style
    // flag should always be set, because even making the Canvas smaller should
    // be followed by a paint event that updates the entire Canvas with new
    // viewport settings.
    : wxGLCanvas(IssiFrame, wxID_ANY, NULL /* attribs */,
                 wxDefaultPosition, wxDefaultSize,
                 wxFULL_REPAINT_ON_RESIZE)
{
	m_IssiFrame = IssiFrame;
	m_StopRender = false;
	m_isTransform = false;
}

// Init world and animation timer

void IssiGLCanvas::InitWorld()
{
	m_World = new World(this, m_IssiFrame);
}

// Paint event

void IssiGLCanvas::OnPaint(wxPaintEvent& WXUNUSED(event))
{
    
	// This is required even though dc is not used otherwise.
 	//wxPaintDC dc(this);

	// Set the OpenGL viewport according to the client size of this Canvas.
	// This is done here rather than in a wxSizeEvent handler because our
	// OpenGL rendering context (and thus viewport setting) is used with
	// multiple Canvases: If we updated the viewport in the wxSizeEvent
	// handler, changing the size of one Canvas causes a viewport setting that
	// is wrong when next another Canvas is repainted.
	//const wxSize ClientSize = GetClientSize();
	//glViewport(0, 0, ClientSize.x, ClientSize.y);

	 // Render the graphics and swap the buffers.
	IssiGLContext& Context = wxGetApp().GetContext(this);
	Context.DrawScene(false, m_World);
	//glFlush();
	SwapBuffers();  
}

// Detects users 3D clicks over canvas

Tool::Idents* IssiGLCanvas::Detect(int x, int y)
{

	GLuint Buff[64] = {0};
 	GLint Hits, View[4];
 
 	// This choose the buffer where store the values for the selection data
 	
 	glSelectBuffer(64, Buff);
 
 	// This retrieve info about the viewport
 	
 	glGetIntegerv(GL_VIEWPORT, View);
 
 	// Switching in selecton mode
 	
 	glRenderMode(GL_SELECT);
 
 	glInitNames();
	glPushName(0);
	
	/*
 		Clearing the name's stack
 		This stack contains all the info about the objects
 	*/
 	
 
 	
 	//	Now fill the stack with one element (or glLoadName will generate an error)
 	
 	
 	
 	//	Now modify the vieving volume, restricting selection area around the cursor
 	
 	glMatrixMode(GL_PROJECTION);
 	glPushMatrix();
 		glLoadIdentity();
 
 		
 		//	restrict the draw to an area around the cursor
 		
 		gluPickMatrix(x, y, 1.0, 1.0, View);
		gluPerspective(60, 1.0, 1.0, 100.0); 
		glMatrixMode(GL_MODELVIEW);
 
 		
 		//	draw only the names in the stack, and fill the array
 		
 		//SwapBuffers();
 		IssiGLContext& Canvas = wxGetApp().GetContext(this);
		Canvas.DrawScene(true, m_World);
   		glMatrixMode(GL_PROJECTION);
 	glPopMatrix();
 
 	/*
 		get number of objects drawed in that area
 		and return to render mode
 	*/
 	Hits = glRenderMode(GL_RENDER);
 
 	
 	//	Print a list of the objects. Reports:
 	
 	Tool::Idents* Idents = new Tool::Idents(Hits);

	Idents->SetIdents(Buff);
 
 	glMatrixMode(GL_MODELVIEW);

	return Idents;

}


// On mouse right clicked

void IssiGLCanvas::OnMouseRDown(wxMouseEvent& event)
{
	const wxPoint pt(event.GetPosition());
	m_World->OnMouseRDown(pt.x, pt.y, Detect(pt.x, GetClientSize().GetHeight() - pt.y), GetClientSize().GetWidth(), GetClientSize().GetHeight());
	
	m_isTransform = false;
	
	if (!(m_World->GetInModeling() || m_World->GetInModelMove()))
	{
		m_ContextMenuNormal = new wxMenu;
		m_ContextMenuNormal->Append(wxID_DELETE, wxT("&Delete object/control point"));
		m_ContextMenuNormal->AppendSeparator();
		m_ContextMenuNormal->Append(wxID_COPY, wxT("&Copy object\tCtrl+C"));
		m_ContextMenuNormal->Append(wxID_PASTE, wxT("&Paste object\tCtrl+V"));
		m_ContextMenuNormal->Append(wxID_CUT, wxT("Cu&t object\tCtrl+X"));
		m_ContextMenuNormal->AppendSeparator();
		m_ContextMenuNormal->Append(wxID_TRANSFORM,wxT("Trans&form Quadric"));
		PopupMenu(m_ContextMenuNormal,pt.x,pt.y);
		delete m_ContextMenuNormal;
	} else Refresh();

}

// On mouse left clicked

void IssiGLCanvas::OnMouseLDown(wxMouseEvent& event)
{
	
	SetFocus();

	wxPoint pt(event.GetPosition());

	m_World->OnMouseLDown(Detect(pt.x,GetClientSize().GetHeight() - pt.y));

	m_isTransform = false;

	Refresh();
}

// On mouse move

void IssiGLCanvas::OnMouseMove(wxMouseEvent& event)
{
	const wxPoint pt(event.GetPosition());
	
	m_World->OnMouseMove(pt.x, pt.y, GetClientSize().GetWidth(), GetClientSize().GetHeight());

	if (event.LeftIsDown() || m_isTransform)
		Refresh();

}

// On mouse left up

void IssiGLCanvas::OnMouseLUp(wxMouseEvent& event)
{
	const wxPoint pt(event.GetPosition());
	
	m_World->OnMouseLUp(pt.x, pt.y);
}

// On mouse right up

void IssiGLCanvas::OnMouseRUp(wxMouseEvent& event)
{
	const wxPoint pt(event.GetPosition());
	
	m_World->OnMouseRUp();

}

// The user moves the mouse wheel. Zooms the imag

void IssiGLCanvas::OnMouseWheel(wxMouseEvent& event)
{
	m_WheelRotation+= event.GetWheelRotation();
	const int Zooming = m_WheelRotation / event.GetWheelDelta();
	m_WheelRotation -= Zooming * event.GetWheelDelta();
		
	m_World->OnMouseWheel(Zooming);

}

// The user press a key

void IssiGLCanvas::OnKeyDown(wxKeyEvent& event)
{
	Refresh();
	if (m_World->OnKeyDown(event.GetKeyCode()))
		event.Skip();
}

// Change size event

void IssiGLCanvas::OnSize(wxSizeEvent& WXUNUSED(event))
{
	const wxSize ClientSize = GetClientSize();
	glViewport(0, 0, ClientSize.x, ClientSize.y);

}

// Menu event reports world

void IssiGLCanvas::OnMenu(int Code)
{
	m_World->OnMenu(Code);
	Refresh();
}

// Menu popup click resports world

void IssiGLCanvas::OnPopupClick(wxCommandEvent &event)
{
	m_World->OnMenu(event.GetId());

	if (event.GetId() != wxID_TRANSFORM)
		Refresh();
	else m_isTransform = true;
}


// Reports IssiControl order

void IssiGLCanvas::OnSetControlOrder(int OrderId)
{
	m_World->OnSetControlOrder(OrderId);
}

// Destroy world

void IssiGLCanvas::DestroyWorld()
{
	delete m_World;
}

// Saves image in a specific format

void IssiGLCanvas::SaveImage(int Id, wxString& Path)
{
	GLint view[4]; 
	glGetIntegerv(GL_VIEWPORT, view); 
	
	// wxImage deletes Bitmap
	void* Bitmap = new unsigned char[3 * view[2] * view[3]];
	
	glPixelStorei(GL_PACK_ALIGNMENT, 1); 
	glReadBuffer(GL_BACK_LEFT); 
	glReadPixels(0, 0, view[2], view[3], GL_RGB, GL_UNSIGNED_BYTE, Bitmap); 
	
	// Put the image into a wxImage 
	wxImage Image(static_cast<int>(view[2]), static_cast<int>(view[3])); 
	Image.SetData(reinterpret_cast<unsigned char*>(Bitmap)); 
	Image = Image.Mirror(false);	
	
	Image.SetMaskFromImage(Image,127,127,127);
	Image.SetMaskColour(0,0,0);
	Image.SetOption(wxIMAGE_OPTION_QUALITY, 100);

	switch(Id)
	{
		case wxID_EXPORT_JPG:
			Image.SaveFile(Path,wxBITMAP_TYPE_JPEG);
			break;
		case wxID_EXPORT_PNG:
			Image.SaveFile(Path,wxBITMAP_TYPE_PNG);
			break;
		case wxID_EXPORT_BMP:
			Image.SaveFile(Path,wxBITMAP_TYPE_BMP);
			break;
	}
}

// Opens XML file

bool IssiGLCanvas::OnXMLFileOpen(wxString& XMLContent)
{
	return m_World->OnXMLFileOpen(XMLContent);	
}

// Saves XML file

bool IssiGLCanvas::OnXMLFileSave(wxString& XMLContent)
{
	return m_World->OnXMLFileSave(XMLContent);	
}

// Export image format

void IssiGLCanvas::OnExport(int Id, wxString& Path)
{
	m_World->OnSaving(true);
	IssiGLContext& Context = wxGetApp().GetContext(this);
	Context.DrawScene(false, m_World);
	glFlush();
	SwapBuffers();
	SaveImage(Id, Path);
	m_World->OnSaving(false);
}

// The users sets 3D text

void IssiGLCanvas::OnText(wxString& Str)
{
	m_World->OnText(Str);
}

// The user changes color

void IssiGLCanvas::OnColor(wxColour& Color)
{
	m_World->OnColor(Color);
}

// The users changes alpha

void IssiGLCanvas::OnAlpha(int Range)
{
	m_World->OnAlpha(Range);
}

// Refresh combo usually when the user drop-downs it

void IssiGLCanvas::OnComboObject(wxComboBox* Combo)
{
	Combo->Clear();
	m_World->OnComboObject(*Combo);
}

// Selects in combo

void IssiGLCanvas::OnComboSelect(wxComboBox* Combo)
{
	wxString Str = Combo->GetStringSelection();
	m_World->OnComboSelect(Str);
}

// Destructor

IssiGLCanvas::~IssiGLCanvas()
{
	DestroyWorld();
}

// ----------------------------------------------------------------------------
// IssiFrame: main application window
// ----------------------------------------------------------------------------

BEGIN_EVENT_TABLE(IssiFrame, wxFrame)
	EVT_MENU(wxID_NEW, IssiFrame::OnNewWindow)
    EVT_MENU(wxID_CLOSE, IssiFrame::OnClose)
	EVT_MENU(wxID_NURBS, IssiFrame::OnNurbs)
	EVT_MENU(wxID_BEZIER, IssiFrame::OnBezier)
	EVT_MENU(wxID_CYLINDER,IssiFrame::OnCylinder)
	EVT_MENU(wxID_SPHERE,IssiFrame::OnSphere)
	EVT_MENU(wxID_CONE,IssiFrame::OnCone)
	EVT_MENU(wxID_TEAPOT,IssiFrame::OnTeapot)
	EVT_MENU(wxID_TETRAHEDRON,IssiFrame::OnTetrahedron)
	EVT_MENU(wxID_OCTAHEDRON,IssiFrame::OnOctahedron)
	EVT_MENU(wxID_DODECAHEDRON,IssiFrame::OnDodecahedron)
	EVT_MENU(wxID_ICOSAHEDRON,IssiFrame::OnIcosahedron)
	EVT_MENU(wxID_ROMBIC_DODECAHEDRON,IssiFrame::OnRombicDodecahedron)
	EVT_MENU(wxID_TORUS,IssiFrame::OnTorus)
	EVT_MENU(wxID_COPY,IssiFrame::OnCopy)
	EVT_MENU(wxID_PASTE,IssiFrame::OnPaste)
	EVT_MENU(wxID_CUT,IssiFrame::OnCut)
	EVT_MENU(wxID_COLOR,IssiFrame::OnColor)
	EVT_MENU(wxID_EXPORT_JPG,IssiFrame::OnExport)
	EVT_MENU(wxID_EXPORT_PNG,IssiFrame::OnExport)
	EVT_MENU(wxID_EXPORT_BMP,IssiFrame::OnExport)
	EVT_MENU(wxID_DELETE,IssiFrame::OnDelete)
	EVT_MENU(wxID_UNDO,IssiFrame::OnUndo)
	EVT_MENU(wxID_REDO,IssiFrame::OnRedo)
	EVT_MENU(wxID_SAVEAS,IssiFrame::OnSaveAs)
	EVT_MENU(wxID_SAVE,IssiFrame::OnSave)
	EVT_MENU(wxID_OPEN,IssiFrame::OnOpen)
	EVT_MENU(wxID_ABOUT,IssiFrame::OnAbout)
	EVT_MENU(wxID_HELP,IssiFrame::OnHelp)
	EVT_MENU(wxID_TEXT, IssiFrame::OnText)
	EVT_SCROLL(IssiFrame::OnSliderScroll)
	EVT_COMBOBOX_DROPDOWN(wxID_COMBO, IssiFrame::OnComboObject)
	EVT_COMBOBOX(wxID_COMBO, IssiFrame::OnComboObject)
END_EVENT_TABLE()

// IssiFrame constructor

IssiFrame::IssiFrame(wxWindow* parent)
       : wxFrame(parent, wxID_ANY, wxT("ISSIGraph Surface Editor"))
{
	
	m_IssiGLCanvas = new IssiGLCanvas(this);

	m_ToolBar = new wxToolBar(this, wxID_TOOLBAR, wxDefaultPosition, wxDefaultSize, wxTB_HORIZONTAL | wxTB_DOCKABLE | wxNO_BORDER | wxTB_FLAT);
	m_ToolBar->SetToolBitmapSize(wxSize(24,24));

	PopulateBitmaps(); 
	PopulateToolBar();
	PopulateMenu();

	SetIcon(wxIcon(appico));
	CreateStatusBar(5);
    SetClientSize(1024, 768);
    Show();

    // test IsDisplaySupported() function:
    static const int attribs[] = { WX_GL_RGBA, WX_GL_DOUBLEBUFFER, 0 };
    SetStatusText(_("Doble-buffer supported: ") + (wxGLCanvas::IsDisplaySupported(attribs) ? _("YES") : _("NOT")),0);
	
	m_IssiGLCanvas->InitWorld();

	(new IssiControl(this, m_IssiGLCanvas))->Show();

}

// Populates bitmaps

void IssiFrame::PopulateBitmaps()
{
	New = wxArtProvider::GetBitmap(wxART_NEW, wxART_TOOLBAR);
	Closebmp = wxArtProvider::GetBitmap(wxART_QUIT, wxART_TOOLBAR);
	Open = wxArtProvider::GetBitmap(wxART_FILE_OPEN, wxART_TOOLBAR);
	Saveas = wxArtProvider::GetBitmap(wxART_FILE_SAVE_AS, wxART_TOOLBAR);
	Save = wxArtProvider::GetBitmap(wxART_FILE_SAVE, wxART_TOOLBAR);
	Copy = wxArtProvider::GetBitmap(wxART_COPY, wxART_TOOLBAR);
	Paste = wxArtProvider::GetBitmap(wxART_PASTE, wxART_TOOLBAR);
	Cut = wxArtProvider::GetBitmap(wxART_CUT, wxART_TOOLBAR);
	Delete = wxArtProvider::GetBitmap(wxART_DELETE, wxART_TOOLBAR);
	Undo = wxArtProvider::GetBitmap(wxART_GOTO_FIRST, wxART_TOOLBAR);
	Redo = wxArtProvider::GetBitmap(wxART_GOTO_LAST, wxART_TOOLBAR);
	Help = wxArtProvider::GetBitmap(wxART_HELP_BOOK, wxART_TOOLBAR);
	Color = wxBitmap(colorico);
	About = wxBitmap(aboutico);
}

// Populates application menu

void IssiFrame::PopulateMenu()
{
	// Make a menubar
    m_MenuMain = new wxMenu;
    
	wxMenuItem* Item = new wxMenuItem(m_MenuMain, wxID_NEW);
	Item->SetBitmap(New);
	m_MenuMain->Append(Item);
	m_MenuMain->AppendSeparator();
    
	Item = new wxMenuItem(m_MenuMain, wxID_CLOSE);
	Item->SetBitmap(Closebmp);
	m_MenuMain->Append(Item);
	m_MenuMain->AppendSeparator();
	
	Item = new wxMenuItem(m_MenuMain, wxID_OPEN, _("&Open XML"));
	Item->SetBitmap(Open);
	m_MenuMain->Append(Item);


	Item = new wxMenuItem(m_MenuMain, wxID_SAVEAS, _("Sa&ve as XML"));
	Item->SetBitmap(Saveas);
	m_MenuMain->Append(Item);

	Item = new wxMenuItem(m_MenuMain, wxID_SAVE, _("&Save XML"));
	Item->SetBitmap(Save);
	m_MenuMain->Append(Item);
	
	Item = new wxMenuItem(m_MenuMain, wxID_ABOUT, _("&About"));
	Item->SetBitmap(About);
	m_MenuMain->Append(Item);

	wxMenu* MenuSurface = new wxMenu;
	MenuSurface->Append(wxID_NURBS,wxT("&NURBS"));
	MenuSurface->Append(wxID_BEZIER,wxT("&BEZIER"));
	MenuSurface->Append(wxID_CYLINDER,wxT("&Cylinder"));
	MenuSurface->Append(wxID_SPHERE,wxT("&Sphere"));
	MenuSurface->Append(wxID_CONE,wxT("Co&ne"));
	MenuSurface->Append(wxID_TEAPOT, wxT("T&eapot"));
	MenuSurface->Append(wxID_TETRAHEDRON, wxT("&Treahedron"));
	MenuSurface->Append(wxID_OCTAHEDRON, wxT("&Ocatahedron"));
	MenuSurface->Append(wxID_DODECAHEDRON, wxT("&Dodecahedron"));
	MenuSurface->Append(wxID_ICOSAHEDRON, wxT("&Icosahedron"));
	MenuSurface->Append(wxID_ROMBIC_DODECAHEDRON, wxT("&Rombic dodecahedron"));
	MenuSurface->Append(wxID_TORUS, wxT("Tor&us"));


	wxMenu* MenuEdit = new wxMenu;
	
	Item = new wxMenuItem(m_MenuMain, wxID_UNDO, _("&Undo"));
	Item->SetBitmap(Undo);
	MenuEdit->Append(Item);
	
	Item = new wxMenuItem(m_MenuMain, wxID_REDO, _("&Redo"));
	Item->SetBitmap(Redo);
	MenuEdit->Append(Item);
	
	MenuEdit->AppendSeparator();

	Item = new wxMenuItem(m_MenuMain, wxID_COPY, _("&Copy\tCtrl+C"));
	Item->SetBitmap(Copy);
	MenuEdit->Append(Item);
	
	Item = new wxMenuItem(m_MenuMain, wxID_PASTE, _("&Paste\tCtrl+V"));
	Item->SetBitmap(Paste);
	MenuEdit->Append(Item);
	
	Item = new wxMenuItem(m_MenuMain, wxID_CUT, _("&Cut\tCtrl+X"));
	Item->SetBitmap(Cut);
	MenuEdit->Append(Item);

	MenuEdit->AppendSeparator();
	
	Item = new wxMenuItem(m_MenuMain, wxID_DELETE, _("&Delete object/control point"));
	Item->SetBitmap(Delete);
	MenuEdit->Append(Item);

	wxMenu* MenuText = new wxMenu;

	MenuText->Append(wxID_TEXT,"&Create text");
	
	wxMenu* MenuColor = new wxMenu;

	MenuColor->Append(wxID_COLOR,"&Add color");

	wxMenu* MenuExport = m_SaveMenu = new wxMenu;
	MenuExport->Append(wxID_EXPORT_JPG,"Export as &JPEG with 100% of quality");
	MenuExport->Append(wxID_EXPORT_PNG,"Export as &PNG");
	MenuExport->Append(wxID_EXPORT_BMP,"Export as &BMP");

	wxMenu* MenuHelp = new wxMenu;
	Item = new wxMenuItem(MenuHelp, wxID_HELP);
	Item->SetBitmap(Help);
	MenuHelp->Append(Item);
	
	
	wxMenuBar* MenuBar = new wxMenuBar;
  	MenuBar->Append(m_MenuMain, wxT("&Application"));
	MenuBar->Append(MenuEdit, wxT("&Edit"));
	MenuBar->Append(MenuSurface, wxT("&Surface"));
	MenuBar->Append(MenuText, wxT("&Text"));
	MenuBar->Append(MenuColor, wxT("&Color"));
    MenuBar->Append(MenuExport, wxT("E&xport Canvas"));
	MenuBar->Append(MenuHelp, wxT("&Help"));

	SetMenuBar(MenuBar);
}

// Populate application toolbar 

void IssiFrame::PopulateToolBar()
{
	
	m_ToolBar->InsertTool(0, wxID_NEW,_(""), New);
	m_ToolBar->SetToolShortHelp(wxID_NEW, _("New window is opened"));
	m_ToolBar->InsertTool(1, wxID_CLOSE, _(""), Closebmp);
	m_ToolBar->SetToolShortHelp(wxID_CLOSE, _("The window is closed"));
	
	m_ToolBar->InsertSeparator(2);
	
	m_ToolBar->InsertTool(3, wxID_OPEN, _(""), Open);
	m_ToolBar->SetToolShortHelp(wxID_OPEN, _("Open XML file"));
	m_ToolBar->InsertTool(4, wxID_SAVEAS, _(""), Saveas);
	m_ToolBar->SetToolShortHelp(wxID_SAVEAS, _("Save AS XML file"));
	m_ToolBar->InsertTool(5, wxID_SAVE, _(""), Save);
	m_ToolBar->SetToolShortHelp(wxID_SAVE, _("Save XML file"));
	
	m_ToolBar->InsertSeparator(6);
	
	m_ToolBar->InsertTool(7, wxID_COPY, _(""), Copy);
	m_ToolBar->SetToolShortHelp(wxID_COPY, _("Copy"));
	m_ToolBar->InsertTool(8, wxID_PASTE, _(""), Paste);
	m_ToolBar->SetToolShortHelp(wxID_PASTE, _("Paste"));
	m_ToolBar->InsertTool(9, wxID_CUT, _(""), Cut);
	m_ToolBar->SetToolShortHelp(wxID_CUT, _("Cut"));
	m_ToolBar->InsertTool(10, wxID_DELETE, _(""), Delete);
	m_ToolBar->SetToolShortHelp(wxID_DELETE, _("Delete object"));
	
	m_ToolBar->InsertSeparator(11);
	
	m_ToolBar->InsertTool(12, wxID_UNDO, _(""), Undo);
	m_ToolBar->SetToolShortHelp(wxID_UNDO, _("Undo"));
	m_ToolBar->InsertTool(13, wxID_REDO, _(""), Redo);
	m_ToolBar->SetToolShortHelp(wxID_REDO, _("Redo"));
	m_ToolBar->InsertSeparator(14);
	m_ToolBar->InsertTool(15, wxID_COLOR, _(""), Color);
	m_ToolBar->SetToolShortHelp(wxID_COLOR, _("Change color"));
	m_ToolBar->InsertSeparator(16);
	m_ToolBar->InsertTool(17, wxID_HELP, _(""), Help);
	m_ToolBar->SetToolShortHelp(wxID_HELP, _("Help"));
	m_ToolBar->InsertSeparator(18);
	wxSlider* Slider = new wxSlider(m_ToolBar, wxID_SLIDER, 100, 0, 100, wxDefaultPosition, 
		wxSize(200,-1));
	m_ToolBar->InsertControl(19, Slider);
	
	m_ToolBar->InsertSeparator(20);
	
	wxArrayString String;
	String.Add(wxT("Coordinate system"));
	m_Combo = new wxComboBox(m_ToolBar, wxID_COMBO, _(""), wxDefaultPosition, wxDefaultSize, String, wxCB_DROPDOWN | wxCB_READONLY); 
	m_ToolBar->InsertControl(21, m_Combo);
	m_ToolBar->Realize();
	SetToolBar(m_ToolBar);
}

// Closes application

void IssiFrame::OnClose(wxCommandEvent& WXUNUSED(event))
{
	// true is to force the frame to close
    Close(true);
}

// New window

void IssiFrame::OnNewWindow( wxCommandEvent& WXUNUSED(event) )
{
	new IssiFrame(this);
}

// The user wants to create a nurbs

void IssiFrame::OnNurbs(wxCommandEvent& WXUNUSED(event))
{
	SetCursor(wxCursor(wxCURSOR_CROSS));
	m_IssiGLCanvas->OnMenu(ID_NURBS);
	
}

// Creates a bezier handler

void IssiFrame::OnBezier(wxCommandEvent& WXUNUSED(event))
{
	SetCursor(wxCursor(wxCURSOR_CROSS));
	m_IssiGLCanvas->OnMenu(ID_BEZIER);
}

// Creates a cylinder handler

void IssiFrame::OnCylinder(wxCommandEvent& WXUNUSED(event))
{
	SetCursor(wxCursor(wxCURSOR_CROSS));
	m_IssiGLCanvas->OnMenu(ID_CYLINDER);
}

// Creates a sphere handler

void IssiFrame::OnSphere(wxCommandEvent& WXUNUSED(event))
{
	SetCursor(wxCursor(wxCURSOR_CROSS));
	m_IssiGLCanvas->OnMenu(ID_SPHERE);
}

// Creates a cone handler

void IssiFrame::OnCone(wxCommandEvent& WXUNUSED(event))
{
	SetCursor(wxCursor(wxCURSOR_CROSS));
	m_IssiGLCanvas->OnMenu(ID_CONE);
}

// Creates teapot handler

void IssiFrame::OnTeapot(wxCommandEvent& WXUNUSED(event))
{
	SetCursor(wxCursor(wxCURSOR_CROSS));
	m_IssiGLCanvas->OnMenu(ID_TEAPOT);
}

// Creates tetrahedron handler

void IssiFrame::OnTetrahedron(wxCommandEvent& WXUNUSED(event))
{
	SetCursor(wxCursor(wxCURSOR_CROSS));
	m_IssiGLCanvas->OnMenu(ID_TETRAHEDRON);
}

// Creates octahedron handler

void IssiFrame::OnOctahedron(wxCommandEvent& WXUNUSED(event))
{
	SetCursor(wxCursor(wxCURSOR_CROSS));
	m_IssiGLCanvas->OnMenu(ID_OCTAHEDRON);
}

// Creates dodecahedron handler

void IssiFrame::OnDodecahedron(wxCommandEvent& WXUNUSED(event))
{
	SetCursor(wxCursor(wxCURSOR_CROSS));
	m_IssiGLCanvas->OnMenu(ID_DODECAHEDRON);
}

// Creates icosahedron handler

void IssiFrame::OnIcosahedron(wxCommandEvent& WXUNUSED(event))
{
	SetCursor(wxCursor(wxCURSOR_CROSS));
	m_IssiGLCanvas->OnMenu(ID_ICOSAHEDRON);
}

// Creates dodecahedron handler

void IssiFrame::OnRombicDodecahedron(wxCommandEvent& WXUNUSED(event))
{
	SetCursor(wxCursor(wxCURSOR_CROSS));
	m_IssiGLCanvas->OnMenu(ID_ROMBIC_DODECAHEDRON);
}

// Creates torus handler

void IssiFrame::OnTorus(wxCommandEvent& WXUNUSED(event))
{
	SetCursor(wxCursor(wxCURSOR_CROSS));
	m_IssiGLCanvas->OnMenu(ID_TORUS);
}

// On copy handler

void IssiFrame::OnCopy(wxCommandEvent& WXUNUSED(event))
{
	m_IssiGLCanvas->OnMenu(wxID_COPY);
}

// On paste handler

void IssiFrame::OnPaste(wxCommandEvent& WXUNUSED(event))
{
	m_IssiGLCanvas->OnMenu(wxID_PASTE);
}

// On cut handler

void IssiFrame::OnCut(wxCommandEvent& WXUNUSED(event))
{
	m_IssiGLCanvas->OnMenu(wxID_CUT);
}

// On delete handler

void IssiFrame::OnDelete(wxCommandEvent& WXUNUSED(event))
{
	m_IssiGLCanvas->OnMenu(wxID_DELETE);
	m_IssiGLCanvas->OnComboObject(m_Combo);
	m_IssiGLCanvas->Refresh();
}

// On undo handler

void IssiFrame::OnUndo(wxCommandEvent& WXUNUSED(event))
{
	m_IssiGLCanvas->OnMenu(ID_UNDO);

}

// On redo handler

void IssiFrame::OnRedo(wxCommandEvent& WXUNUSED(event))
{
	m_IssiGLCanvas->OnMenu(ID_REDO);
}

// The uses wants to create 3D text

void IssiFrame::OnText(wxCommandEvent& WXUNUSED(event))
{
	wxString Str;
	Str = ::wxGetTextFromUser(_("Input text to render:"), _("Input text"), _(""), this);
	
	if (!Str.IsEmpty())
	{
		m_IssiGLCanvas->OnText(Str);
		m_IssiGLCanvas->OnComboObject(m_Combo);
		m_IssiGLCanvas->Refresh();
	}
}

// Change color handler

void IssiFrame::OnColor(wxCommandEvent& WXUNUSED(event))
{
	wxColourData Data;
	Data.SetChooseFull(true);
	
	for (int i = 0; i < 16; i++)
	{
		wxColour Color(i*16, i*16, i*16);
		Data.SetCustomColour(i, Color);
	}
	
	wxColourDialog ColorDlg(this, &Data);
	
	if (ColorDlg.ShowModal() == wxID_OK)
	{
		wxColour Color = ColorDlg.GetColourData().GetColour();
		m_IssiGLCanvas->OnColor(Color);		
	}
	m_IssiGLCanvas->Refresh();

}

// Save as XML handler

void IssiFrame::OnSaveAs(wxCommandEvent& WXUNUSED(event))
{
	wxString XMLContent;
	wxTextFile XMLFile;
	wxFileDialog Dialog(this,wxT("Save XML ISSI format as "), wxT("."),wxT("file"), " (*.xml)|*.xml", wxFD_SAVE);
	
	if (Dialog.ShowModal() == wxID_OK) // everything is ok...
	{
		XMLFile.Create(m_SavedXML = Dialog.GetPath()); // Create file
		XMLFile.AddLine(_("<?xml version=\"1.0\" encoding=\"UTF-8\" ?>")); // XML header tag
		XMLFile.AddLine(_("<issiroot version=\"1.0\">"));
		if (!m_IssiGLCanvas->OnXMLFileSave(XMLContent))
			m_MenuMain->Enable(wxID_SAVE, false);
		else m_MenuMain->Enable(wxID_SAVE, true);

		XMLFile.AddLine(XMLContent);
		XMLFile.AddLine(_("</issiroot>")); // close tag
		XMLFile.Write();
	}
	
}

// Sabe XML handler

void IssiFrame::OnSave(wxCommandEvent& WXUNUSED(event))
{
	wxString XMLContent;
	wxTextFile XMLFile;

	if (!m_SavedXML.IsEmpty()) // It has a filled wxString
	{

		XMLFile.Create(m_SavedXML); // Create file
		XMLFile.AddLine(_("<?xml version=\"1.0\" encoding=\"UTF-8\" ?>"));
		XMLFile.AddLine(_("<issiroot version=\"1.0\">"));
	
		m_IssiGLCanvas->OnXMLFileSave(XMLContent);

		XMLFile.AddLine(XMLContent);
		XMLFile.AddLine(_("</issiroot>"));
		XMLFile.Write();
	} else wxMessageBox("Sorry, nothing to save");

}

// XML open handler

void IssiFrame::OnOpen(wxCommandEvent& WXUNUSED(event))
{
	wxString XMLContent;
	wxTextFile XMLFile;
	wxFileDialog Dialog(this,wxT("Open XML ISSI format"), wxT("."),wxT("file"), " (*.xml)|*.xml", wxFD_OPEN);
	
	if (Dialog.ShowModal() == wxID_OK) // if everyting is ok...
	{
		if (XMLFile.Open(m_SavedXML = Dialog.GetPath())) // Get selected path
		{
			XMLContent = XMLFile.GetFirstLine();
			
			if (XMLContent.IsEmpty())
			{
				wxMessageBox("I'm sorry, but the file is empty");
				Refresh();
				return;
			}

			// read all lines one by one
			// until the end of the file
			while(!XMLFile.Eof())
			{
				XMLContent+=XMLFile.GetNextLine();
			}
		
			if (!m_IssiGLCanvas->OnXMLFileOpen(XMLContent))
				m_MenuMain->Enable(wxID_SAVE, false);
			else m_MenuMain->Enable(wxID_SAVE, true);

			Refresh();
		}
		
	}
	
	
}

// Export JPG, PNG or BMP handler

void IssiFrame::OnExport(wxCommandEvent& event)
{
	wxString Wildcard;

	switch(event.GetId())
	{
		case wxID_EXPORT_JPG:
			Wildcard = wxT("JPEG files (*.jpg)|*.jpg");
			break;		
		case wxID_EXPORT_PNG:
			Wildcard = wxT("PNG files (*.png)|*.png");
			break;		
		case wxID_EXPORT_BMP:
			Wildcard = wxT("BMP files (*.bmp)|*.bmp");
			break;		
	}

	wxFileDialog Dialog(this,wxT("Save a image file as ") + Wildcard, wxT("."),wxT("image"), Wildcard, wxFD_SAVE);
	
	if (Dialog.ShowModal() == wxID_OK)
	{
		wxString Str = Dialog.GetPath();
		m_IssiGLCanvas->OnExport(event.GetId(), Str);
	}
}

// On about handler

void IssiFrame::OnAbout(wxCommandEvent& WXUNUSED(event))
{
	IssiAbout AboutDlg(this);

	AboutDlg.ShowModal();
}

// On help handler

void IssiFrame::OnHelp(wxCommandEvent& WXUNUSED(event))
{
	IssiHelp HelpDlg(this);

	HelpDlg.ShowModal();
}

// Alpha channel handler

void IssiFrame::OnSliderScroll(wxScrollEvent& event) {
     int Alpha = event.GetPosition();

	 m_IssiGLCanvas->OnAlpha(Alpha);
	 m_IssiGLCanvas->Refresh();

}

// The user changes combo

void IssiFrame::OnComboObject(wxCommandEvent& event)
{
	if (event.GetEventType() == wxEVT_COMMAND_COMBOBOX_SELECTED) // It's a selection
		m_IssiGLCanvas->OnComboSelect(m_Combo);
	else // otherwise (simple drop-down). Refresh combo.
		m_IssiGLCanvas->OnComboObject(m_Combo);
	m_IssiGLCanvas->Refresh();
}


