/*****************************************************************************
   ISSIGraph: Experimental Surface Editor Prototype						 
   Copyright (C) 2013 by Carlos Jimnez de Parga								 
   
   Trabajo para el:
   MASTER EN INVESTIGACIN EN INGENIERA DE SOFTWARE Y SISTEMAS INFORMTICOS
			REPRESENTACIN GRFICA DE SUPERFICIES IMPLCITAS			
	DEPARTAMENTO DE INGENIERA DE SOFTWARE Y SISTEMAS INFORMTICOS 
							- UNED -

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
******************************************************************************/
/* 
	FILE: Main.h
	PURPOSE: User interface (header)
	Remarks: This file (main.h) is based on the Cube.h wxWidget 2.9.3 OpenGL example by Julian Smart (c)
	Author: Carlos Jimnez de Parga (c)
	License: GPL v3.0 and wxWindows
*/

#ifndef _MAIN_H_
#define _MAIN_H_

#include "gl/glew.h"
#include "wx/glCanvas.h"
#include "wx/image.h"
#include "wx/filedlg.h"
#include "wx/textfile.h"
#include "wx/xrc/xmlres.h"
#include "wx/bannerwindow.h"
#include "wx/sizer.h"
#include "wx/slider.h"
#include "wx/colordlg.h"
#include "wx/html/htmlwin.h"
#include "wx/combobox.h"
#include "world.h"
#include "definitions.h"

// ----------------------------------------------------------------------------
// IssiGLContext class
// The rendering context used by all GL Canvases
// ----------------------------------------------------------------------------

class IssiGLContext : public wxGLContext
{
public:
    IssiGLContext(wxGLCanvas* Canvas);
	void DrawScene(bool Lactive, World* World);
};


// ----------------------------------------------------------------------------
// IssiLogo
// Displays start-up banner
// ----------------------------------------------------------------------------

class IssiLogo : public wxDialog
{
public:
	IssiLogo();
	void OnTimer(wxTimerEvent& WXUNUSED(event));
private:
	wxBannerWindow m_Banner;
	wxGauge* m_Gauge;
	wxTimer m_Timer;
	int m_TimerValue;
	DECLARE_EVENT_TABLE()
};

// ----------------------------------------------------------------------------
// IssiAbout
// Displays About dialog box
// ----------------------------------------------------------------------------

class IssiAbout : public wxDialog
{
public:
	IssiAbout(wxFrame* Parent);
private:
	wxBannerWindow m_Banner;
};

// ----------------------------------------------------------------------------
// IssiHelp
// Displays Help dialog box
// ----------------------------------------------------------------------------


class IssiHelp : public wxDialog
{
public:
	IssiHelp(wxFrame* Parent);
};

// ----------------------------------------------------------------------------
// IssiApp
// Application manager
// ----------------------------------------------------------------------------

class IssiApp : public wxApp
{
public:
    IssiApp() { m_glContext = NULL; }

    // Returns the shared context used by all frames and sets it as current for
    // the given Canvas.
    IssiGLContext& GetContext(wxGLCanvas* Canvas);
	// virtual wxApp methods
    virtual bool OnInit();
    virtual int OnExit();

private:
    // the GL context we use for all our windows
    IssiGLContext* m_glContext;
};


// ----------------------------------------------------------------------------
// IssiGLCanvas class
// Handles the frame's owner canvas
// ----------------------------------------------------------------------------

class IssiGLCanvas : public wxGLCanvas
{
public:
    IssiGLCanvas(wxFrame* IssiFrame);
	~IssiGLCanvas();
	void InitWorld();
	void OnMenu(int Code);
	void OnExport(int Id, wxString& Path);
	bool OnXMLFileSave(wxString& XMLContent);
	bool OnXMLFileOpen(wxString& XMLContent);
	void OnColor(wxColour& Color);
	void OnAlpha(int Range);
	void OnText(wxString& Text);
	void OnComboObject(wxComboBox* Combo);
	void OnComboSelect(wxComboBox* Combo);
	void OnSetControlOrder(int OrderId);
	void DestroyWorld();
private:
	void OnPaint(wxPaintEvent& event);
    void OnMouseRDown(wxMouseEvent& event);
	void OnMouseLDown(wxMouseEvent& event);
	void OnMouseMove(wxMouseEvent& event);
	void OnMouseLUp(wxMouseEvent& event);
	void OnMouseRUp(wxMouseEvent& event);
	void OnMouseWheel(wxMouseEvent& event);
	void Spin(float xSpin, float ySpin);
	void OnSize(wxSizeEvent& WXUNUSED(event));
	void OnPopupClick(wxCommandEvent &event);
	void OnKeyDown(wxKeyEvent &event);
	void SaveImage(int Id, wxString& Path);
	Tool::Idents* Detect(int x, int y);
	
	World* m_World; // The owner World object
	int m_WheelRotation;
	bool m_StopRender; // If saving
	bool m_isTransform;
	wxMenu* m_ContextMenuNormal; // Context menu object
	wxFrame* m_IssiFrame;
	DECLARE_EVENT_TABLE()
};


// ----------------------------------------------------------------------------
// IssiControl class
// Handles the 3D control dialog
// ----------------------------------------------------------------------------

class IssiControl : public wxDialog
{
public:
	IssiControl(wxFrame* Parent, IssiGLCanvas* Canvas);
private:
	void OnZoomMinus(wxCommandEvent& event);
	void OnZoomPlus(wxCommandEvent& event);
	void OnLeft(wxCommandEvent& event);
	void OnRight(wxCommandEvent& event);
	void OnUp(wxCommandEvent& event);
	void OnDown(wxCommandEvent& event);
	void OnCenter(wxCommandEvent& event);
	void OnRotateYPlus(wxCommandEvent& event);
	void OnRotateYMinus(wxCommandEvent& event);
	
	IssiGLCanvas*	m_Canvas;
	wxBitmapButton* m_ButtonZoomMinus;
	wxBitmapButton* m_ButtonUp;
	wxBitmapButton* m_ButtonZoomPlus;
	wxBitmapButton* m_ButtonLeft; 
	wxBitmapButton* m_ButtonCenter;
	wxBitmapButton* m_ButtonRight;
	wxBitmapButton* m_ButtonRotateYMinus;
	wxBitmapButton* m_ButtonDown;
	wxBitmapButton* m_ButtonRotateYPlus;
	DECLARE_EVENT_TABLE()
};


// ----------------------------------------------------------------------------
// IssiFrame class
// The main frame for the canvas
// ----------------------------------------------------------------------------

class IssiFrame : public wxFrame
{
public:
    IssiFrame(wxWindow* parent);
private:
	void OnClose(wxCommandEvent& event);
    void OnNewWindow(wxCommandEvent& event);
	void OnNurbs(wxCommandEvent& event);
	void OnBezier(wxCommandEvent& event);
	void OnCylinder(wxCommandEvent& event);
    void OnSphere(wxCommandEvent& event);
	void OnCone(wxCommandEvent& event);
	void OnTeapot(wxCommandEvent& event);
	void OnTetrahedron(wxCommandEvent& event);
	void OnOctahedron(wxCommandEvent& event);
	void OnDodecahedron(wxCommandEvent& event);
	void OnIcosahedron(wxCommandEvent& event);
	void OnRombicDodecahedron(wxCommandEvent& event);
	void OnTorus(wxCommandEvent& event);
	void OnCopy(wxCommandEvent& event);
	void OnCut(wxCommandEvent& event);
	void OnPaste(wxCommandEvent& event);
	void OnText(wxCommandEvent& event);
	void OnExport(wxCommandEvent& event);
	void OnDelete(wxCommandEvent& event);
	void OnUndo(wxCommandEvent& event);
	void OnRedo(wxCommandEvent& event);
	void OnSaveAs(wxCommandEvent& event);
	void OnSave(wxCommandEvent& event);
	void OnOpen(wxCommandEvent& event);
	void OnAbout(wxCommandEvent& event);
	void OnColor(wxCommandEvent& event);
	void OnHelp(wxCommandEvent& event);
	void OnSliderScroll(wxScrollEvent& event);
	void OnComboObject(wxCommandEvent& event);
	void OnReleaseTB(wxCommandEvent& event);
	void PopulateBitmaps();
	void PopulateToolBar();
	void PopulateMenu();

	IssiGLCanvas* m_IssiGLCanvas;
	wxMenu* m_SaveMenu;
	wxMenu* m_MenuMain;
	wxString m_SavedXML;
	wxToolBar* m_ToolBar;
	wxComboBox* m_Combo;

	wxBitmap New; 
	wxBitmap Closebmp; 
	wxBitmap Open; 
	wxBitmap Saveas; 
	wxBitmap Save; 
	wxBitmap Copy; 
	wxBitmap Paste; 
	wxBitmap Cut; 
	wxBitmap Delete; 
	wxBitmap Undo;
	wxBitmap Redo;
	wxBitmap Help;
	wxBitmap About;
	wxBitmap Color;

	DECLARE_EVENT_TABLE()
};




#endif // _MAIN_H_
