/* 
	FILE: Quadric.cpp
	PURPOSE: Quadrics and basic 3D objects (implementation)
	Author: Carlos Jimnez de Parga (c)
	License: GPL v3.0
*/

#include "quadric.h"

// ----------------------------------------------------------------------------
// Quadric implementation
// ----------------------------------------------------------------------------

Quadric::Quadric()
{
	qobj = gluNewQuadric();
}

Quadric::Quadric(float ptx, float pty,float ptz,float width, float height):QuadricBox(ptx, pty, ptz, width, height)
{
	qobj = gluNewQuadric(); // Generate Quadric
}

void Quadric::Copy(Quadric* q)
{
	QuadricBox::Copy(q);
}


// Change coordinates

void Quadric::ChangeCoords(float ptx, float pty, float ptz)
{
	m_tx = m_x = ptx;
	m_ty = m_y = pty;
	m_tz = m_z = ptz;
}

// Draw move handler

void Quadric::DrawSelection()
{
	glPushMatrix();
	glColor3f(1.0f ,0.5f,0.5f);
	
	glTranslatef(m_tx,m_ty,m_tz);
	
	glBegin(GL_QUADS);
		glVertex3f(0.0f, 0.0f, 0.0f);
		glVertex3f(2.0f, 0.0f, 0.0f);
		glVertex3f(2.0f, 2.0f, 0.0f);
		glVertex3f(0.0f, 2.0f, 0.0f);
	glEnd();
	
	glPopMatrix();
}


// Destroy quadric

void Quadric::Destroy()
{
	if (qobj) gluDeleteQuadric(qobj);
}

Quadric::~Quadric() 
{ 
	Destroy();
}


// ----------------------------------------------------------------------------
// TwoDimension class implementation
// ----------------------------------------------------------------------------

void TwoDimension::ChangeDimensions(float Radius, float Height)
{
	m_w = Radius;
	m_h = Height;
}

void TwoDimension::Copy(TwoDimension* Two)
{
	Two->DoCopy(m_ScaleW, m_ScaleH);
	Quadric::Copy(Two);
}

void TwoDimension::DoCopy(float ScaleW, float ScaleH)
{
	m_ScaleW = ScaleW;
	m_ScaleH = ScaleH;
}

// Draw selection box

void TwoDimension::DrawBox()
{
	
	glPushMatrix();
	
	glEnable(GL_LINE_STIPPLE);
	glLineStipple(5,0xaaaa);
	
	glColor3f(1.0f,1.0f,0.0f);
	glTranslatef(m_tx,m_ty,m_tz);
	glBegin(GL_QUADS);
		glVertex3f(-m_w, 0.0f, 0.0f);
		glVertex3f(m_w, 0.0f, 0.0f);
		glVertex3f(m_w, m_h, 0.0f);
		glVertex3f(-m_w, m_h, 0.0f);
	glEnd();
		
	glDisable(GL_LINE_STIPPLE); 
	
	glPopMatrix();

}

// ----------------------------------------------------------------------------
// One dimension class implementation
// ----------------------------------------------------------------------------

void OneDimension::ChangeDimensions(float Radius)
{
	m_h = m_w = Radius;
}

// Draw selection box

void OneDimension::DrawBox()
{
	
	glPushMatrix();
	
	glEnable(GL_LINE_STIPPLE);
	glLineStipple(5,0xaaaa);
	
	glColor3f(1.0f,1.0f,0.0f);
	glTranslatef(m_tx,m_ty,m_tz);
	glBegin(GL_QUADS);
		glVertex3f(-m_w, -m_h, 0.0f);
		glVertex3f(m_w, -m_h, 0.0f);
		glVertex3f(m_w, m_h, 0.0f);
		glVertex3f(-m_w, m_h, 0.0f);
	glEnd();
		
	glDisable(GL_LINE_STIPPLE); 
	
	glPopMatrix();

}


// ----------------------------------------------------------------------------
// Cylinder implementation
// ----------------------------------------------------------------------------

Cylinder::Cylinder(GLuint Name)
{
	m_Name = Name;
}


// Draws the cylinder

void Cylinder::Draw()
{
	glColor4f(m_Color.Red() / 255.0f ,m_Color.Green() / 255.0f, m_Color.Blue() / 255.0f, m_Alpha/ 100.0f);
	
	glPushMatrix();
	
	glTranslatef(m_tx ,m_ty,m_tz);
	glRotatef(m_xAngle,1.0f,0.0f,0.0f);
	glRotatef(m_yAngle,0.0f,1.0f,0.0f);
	glLoadName(m_Name);		
	gluCylinder(qobj, m_w, m_w, m_h, 30, 30);
	glPopMatrix();
	
}

void Cylinder::Save(wxString& XMLContent)
{
	XMLContent+=_("<issiform form = \"Cylinder\">");
	QuadricBox::Save(XMLContent);
	XMLContent+=_("</issiform>");

}

// ----------------------------------------------------------------------------
// Sphere implementation
// ----------------------------------------------------------------------------

Sphere::Sphere(GLuint Name)
{
	m_Name = Name;
}

void Sphere::Save(wxString& XMLContent)
{
	XMLContent+=_("<issiform form = \"Sphere\">");
	Quadric::Save(XMLContent);
	XMLContent+=_("</issiform>");

}


// Draws the sphere

void Sphere::Draw()
{
	glColor4f(m_Color.Red() / 255.0f ,m_Color.Green() / 255.0f, m_Color.Blue() / 255.0f, m_Alpha/ 100.0f);
	
	glPushMatrix();
	
	glTranslatef(m_tx ,m_ty,m_tz);
	glRotatef(m_xAngle,1.0f,0.0f,0.0f);
	glRotatef(m_yAngle,0.0f,1.0f,0.0f);
	glLoadName(m_Name);		
	gluSphere(qobj, m_w, 30, 30);
	glPopMatrix();
	
}

// ----------------------------------------------------------------------------
// Cone implementation
// ----------------------------------------------------------------------------

Cone::Cone(GLuint Name)
{
	m_Name = Name;	
}

// Draws the cone

void Cone::Draw()
{
	glColor4f(m_Color.Red() / 255.0f ,m_Color.Green() / 255.0f, m_Color.Blue() / 255.0f, m_Alpha/ 100.0f);
	
	glPushMatrix();
	
	glTranslatef(m_tx ,m_ty,m_tz);
	glRotatef(m_xAngle,1.0f,0.0f,0.0f);
	glRotatef(m_yAngle,0.0f,1.0f,0.0f);
	glLoadName(m_Name);		
	gluCylinder(qobj, m_w, 0, m_h, 30, 30);
	glPopMatrix();
	
}

void Cone::Save(wxString& XMLContent)
{
	XMLContent+=_("<issiform form = \"Cone\">");
	QuadricBox::Save(XMLContent);
	XMLContent+=_("</issiform>");

}


// ----------------------------------------------------------------------------
// Teapot implementation
// ----------------------------------------------------------------------------

Teapot::Teapot(GLuint Name)
{
	m_Name = Name;
}

// Draws teapot

void Teapot::Draw()
{
	glColor4f(m_Color.Red() / 255.0f ,m_Color.Green() / 255.0f, m_Color.Blue() / 255.0f, m_Alpha/ 100.0f);
	
	glPushMatrix();
	
	glTranslatef(m_tx ,m_ty,m_tz);
	glRotatef(m_xAngle,1.0f,0.0f,0.0f);
	glRotatef(m_yAngle,0.0f,1.0f,0.0f);
	glLoadName(m_Name);		
	glutWireTeapot(m_w);
	glPopMatrix();
}	

void Teapot::Save(wxString& XMLContent)
{
	XMLContent+=_("<issiform form = \"Teapot\">");
	QuadricBox::Save(XMLContent);
	XMLContent+=_("</issiform>");

}


// ----------------------------------------------------------------------------
// Tetrahedron implementation
// ----------------------------------------------------------------------------

Tetrahedron::Tetrahedron(GLuint Name)
{
	m_Name = Name;
}

// Draws tetrahedron

void Tetrahedron::Draw()
{
	glColor4f(m_Color.Red() / 255.0f ,m_Color.Green() / 255.0f, m_Color.Blue() / 255.0f, m_Alpha/ 100.0f);
	
	glPushMatrix();
	
	glTranslatef(m_tx ,m_ty,m_tz);
	glRotatef(m_xAngle,1.0f,0.0f,0.0f);
	glRotatef(m_yAngle,0.0f,1.0f,0.0f);
	glScalef(m_w, m_w, m_w);
	glLoadName(m_Name);		
	glutWireTetrahedron();
	glPopMatrix();
}	

void Tetrahedron::Save(wxString& XMLContent)
{
	XMLContent+=_("<issiform form = \"Tetrahedron\">");
	QuadricBox::Save(XMLContent);
	XMLContent+=_("</issiform>");

}

// ----------------------------------------------------------------------------
// Octahedron implementation
// ----------------------------------------------------------------------------

Octahedron::Octahedron(GLuint Name)
{
	m_Name = Name;
}

// Draws octahedron

void Octahedron::Draw()
{
	glColor4f(m_Color.Red() / 255.0f ,m_Color.Green() / 255.0f, m_Color.Blue() / 255.0f, m_Alpha/ 100.0f);
	
	glPushMatrix();
	
	glTranslatef(m_tx ,m_ty,m_tz);
	glRotatef(m_xAngle,1.0f,0.0f,0.0f);
	glRotatef(m_yAngle,0.0f,1.0f,0.0f);
	glScalef(m_w, m_w, m_w);
	glLoadName(m_Name);		
	glutWireOctahedron();
	glPopMatrix();
}	

void Octahedron::Save(wxString& XMLContent)
{
	XMLContent+=_("<issiform form = \"Octahedron\">");
	QuadricBox::Save(XMLContent);
	XMLContent+=_("</issiform>");

}


// ----------------------------------------------------------------------------
// Dodecahedron implementation
// ----------------------------------------------------------------------------

Dodecahedron::Dodecahedron(GLuint Name)
{
	m_Name = Name;
}

// Draws dodecahedron

void Dodecahedron::Draw()
{
	glColor4f(m_Color.Red() / 255.0f ,m_Color.Green() / 255.0f, m_Color.Blue() / 255.0f, m_Alpha/ 100.0f);
	
	glPushMatrix();
	
	glTranslatef(m_tx ,m_ty,m_tz);
	glRotatef(m_xAngle,1.0f,0.0f,0.0f);
	glRotatef(m_yAngle,0.0f,1.0f,0.0f);
	glScalef(m_w, m_w, m_w);
	glLoadName(m_Name);		
	glutWireDodecahedron();
	glPopMatrix();
}	

void Dodecahedron::Save(wxString& XMLContent)
{
	XMLContent+=_("<issiform form = \"Dodecahedron\">");
	QuadricBox::Save(XMLContent);
	XMLContent+=_("</issiform>");

}


// ----------------------------------------------------------------------------
// Icosahedron implementation
// ----------------------------------------------------------------------------

Icosahedron::Icosahedron(GLuint Name)
{
	m_Name = Name;
}

// Draws icosahedron

void Icosahedron::Draw()
{
	glColor4f(m_Color.Red() / 255.0f ,m_Color.Green() / 255.0f, m_Color.Blue() / 255.0f, m_Alpha/ 100.0f);
	
	glPushMatrix();
	
	glTranslatef(m_tx ,m_ty,m_tz);
	glRotatef(m_xAngle,1.0f,0.0f,0.0f);
	glRotatef(m_yAngle,0.0f,1.0f,0.0f);
	glScalef(m_w, m_w, m_w);
	glLoadName(m_Name);		
	glutWireIcosahedron();
	glPopMatrix();
}	

void Icosahedron::Save(wxString& XMLContent)
{
	XMLContent+=_("<issiform form = \"Icosahedron\">");
	QuadricBox::Save(XMLContent);
	XMLContent+=_("</issiform>");

}


// ----------------------------------------------------------------------------
// Rombic dodecahedron implementation
// ----------------------------------------------------------------------------

RombicDodecahedron::RombicDodecahedron(GLuint Name)
{
	m_Name = Name;
}

// Draws rombic dodecahedron

void RombicDodecahedron::Draw()
{
	glColor4f(m_Color.Red() / 255.0f ,m_Color.Green() / 255.0f, m_Color.Blue() / 255.0f, m_Alpha/ 100.0f);
	
	glPushMatrix();
	
	glTranslatef(m_tx ,m_ty,m_tz);
	glRotatef(m_xAngle,1.0f,0.0f,0.0f);
	glRotatef(m_yAngle,0.0f,1.0f,0.0f);
	glScalef(m_w, m_w, m_w);
	glLoadName(m_Name);		
	glutWireRhombicDodecahedron();
	glPopMatrix();
}	

void RombicDodecahedron::Save(wxString& XMLContent)
{
	XMLContent+=_("<issiform form = \"RombicDodecahedron\">");
	QuadricBox::Save(XMLContent);
	XMLContent+=_("</issiform>");

}


// ----------------------------------------------------------------------------
// Torus implementation
// ----------------------------------------------------------------------------

Torus::Torus(GLuint Name)
{
	m_Name = Name;
}

// Draw torus

void Torus::Draw()
{
	glColor4f(m_Color.Red() / 255.0f ,m_Color.Green() / 255.0f, m_Color.Blue() / 255.0f, m_Alpha/ 100.0f);
	
	glPushMatrix();
	
	glTranslatef(m_tx ,m_ty,m_tz);
	glRotatef(m_xAngle,1.0f,0.0f,0.0f);
	glRotatef(m_yAngle,0.0f,1.0f,0.0f);
	glLoadName(m_Name);		
	glutWireTorus(m_ScaleW,m_ScaleH,10,10);
	glPopMatrix();
}	


// Draw selection box

void Torus::DrawBox()
{
	
	glPushMatrix();
	
	glEnable(GL_LINE_STIPPLE);
	glLineStipple(5,0xaaaa);
	
	glColor3f(1.0f,1.0f,0.0f);
	glTranslatef(m_tx,m_ty,m_tz);
	glBegin(GL_QUADS);
		glVertex3f(-m_w, -m_h, 0.0f);
		glVertex3f(m_w, -m_h, 0.0f);
		glVertex3f(m_w, m_h, 0.0f);
		glVertex3f(-m_w, m_h, 0.0f);
	glEnd();
		
	glDisable(GL_LINE_STIPPLE); 
	
	glPopMatrix();

}


void Torus::ChangeDimensions(float radius, float height)
{
	 m_ScaleW = radius * 0.5f;
	 m_ScaleH = height * 0.5f;

	 m_h = m_w = m_ScaleW + m_ScaleH;

}

void Torus::Save(wxString& XMLContent)
{
	XMLContent+=_("<issiform form = \"Torus\">");
	XMLContent+=_("<m_ScaleW>") + Tool::Math::ConvertNumber(m_ScaleW) + _("</m_ScaleW>"); 
	XMLContent+=_("<m_ScaleH>") + Tool::Math::ConvertNumber(m_ScaleH) + _("</m_ScaleH>"); 
	QuadricBox::Save(XMLContent);
	XMLContent+=_("</issiform>");

}

void Torus::Load(xml_node<>* Node)
{
	m_ScaleW = atof(Node->value());
	Node = Node->next_sibling();

	m_ScaleH = atof(Node->value());
	Node = Node->next_sibling();

	QuadricBox::Load(Node->first_node());
}



// ----------------------------------------------------------------------------
// Text 3D implementation
// ----------------------------------------------------------------------------

Text::Text(GLuint Name)
{
	m_Name = Name;
}

// Draw text

void Text::Draw()
{
	glColor4f(m_Color.Red() / 255.0f ,m_Color.Green() / 255.0f, m_Color.Blue() / 255.0f, m_Alpha/ 100.0f);
	
	glPushMatrix();
	
	glTranslatef(m_tx ,m_ty,m_tz);
	glRotatef(m_xAngle,1.0f,0.0f,0.0f);
	glRotatef(m_yAngle,0.0f,1.0f,0.0f);
	glScalef(m_ScaleW, m_ScaleH, m_ScaleW);
	glLoadName(m_Name);		
	glutStrokeString(GLUT_STROKE_ROMAN, reinterpret_cast<unsigned const char*>(m_Str.data().AsChar()));
	glPopMatrix();
}	

void Text::Save(wxString& XMLContent)
{
	XMLContent+=_("<issiform form = \"Text\">");
	XMLContent+=_("<m_Str>") + m_Str + _("</m_Str>");
	XMLContent+=_("<m_ScaleW>") + Tool::Math::ConvertNumber(m_ScaleW) + _("</m_ScaleW>"); 
	XMLContent+=_("<m_ScaleH>") + Tool::Math::ConvertNumber(m_ScaleH) + _("</m_ScaleH>"); 
	QuadricBox::Save(XMLContent);
	XMLContent+=_("</issiform>");

}

void Text::SetText(wxString Str)
{
	m_Str = Str;
}

void Text::ChangeDimensions(float radius, float height)
{
	m_w = glutStrokeLength(GLUT_STROKE_ROMAN, reinterpret_cast<unsigned const char*>(m_Str.data().AsChar())) * 0.999f * radius * 0.01;
	m_h = height;
	m_ScaleW = radius * 0.01;
	m_ScaleH = height * 0.01;
}

void Text::Load(xml_node<>* Node)
{
	m_Str = Node->value();
	Node = Node->next_sibling();

	m_ScaleW = atof(Node->value());
	Node = Node->next_sibling();

	m_ScaleH = atof(Node->value());
	Node = Node->next_sibling();

	QuadricBox::Load(Node->first_node());
}

// Performs copy

void Text::Copy(Text* t)
{
	t->DoCopy(m_Str);
	TwoDimension::Copy(t);
}

void Text::DoCopy(wxString& Str)
{
	m_Str = Str;
}

void Text::DrawBox()
{
	glPushMatrix();
	
	glEnable(GL_LINE_STIPPLE);
	glLineStipple(5,0xaaaa);
	
	glColor3f(1.0f,1.0f,0.0f);
	glTranslatef(m_tx,m_ty,m_tz);

	glBegin(GL_QUADS);
		glVertex3f(0.0f, 0.0f, 0.0f);
		glVertex3f(0.0f, m_h, 0.0f);
		glVertex3f(m_w, m_h, 0.0f);
		glVertex3f(m_w, 0.0f, 0.0f);
	glEnd();

	glDisable(GL_LINE_STIPPLE); 
	glPopMatrix();
}