/*****************************************************************************
   ISSIGraph: Experimental Surface Editor Prototype						 
   Copyright (C) 2013 by Carlos Jimnez de Parga								 
   
   Trabajo para el:
   MASTER EN INVESTIGACIN EN INGENIERA DE SOFTWARE Y SISTEMAS INFORMTICOS
			REPRESENTACIN GRFICA DE SUPERFICIES IMPLCITAS			
	DEPARTAMENTO DE INGENIERA DE SOFTWARE Y SISTEMAS INFORMTICOS 
							- UNED -

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
   *************************************************************************/


/* 
	FILE: Quadric.h
	PURPOSE: Quadrics and basic 3D objects (header)
	Author: Carlos Jimnez de Parga (c)
	License: GPL v3.0
*/

#ifndef _QUADRIC_H_
#define _QUADRIC_H_

#include "box.h"
#include "GL/freeglut.h"

// ----------------------------------------------------------------------------
// Quadric class
// ----------------------------------------------------------------------------

class Quadric : public QuadricBox
{
public:
	Quadric();
	Quadric(float ptx, float pty,float ptz,float width, float height);
	~Quadric();
	void ChangeCoords(float ptx, float pty, float ptz); 
	void DrawSelection();
	void Rotate(float xAngle, float yAngle);
	void Copy(Quadric* q);
	void Destroy();
protected:	
	GLUquadric* qobj;
};

// Class for two dimensions quadrics for example: Torus and Text

class TwoDimension : public Quadric
{
public:	
	TwoDimension() { m_ScaleW = 0.0f; m_ScaleH = 0.0f;}
	virtual void ChangeDimensions(float Radius, float Height);
	void DrawBox();
	void Copy(TwoDimension* Two);
protected:
	void DoCopy(float ScaleW, float ScaleH);
	float m_ScaleW, m_ScaleH;
};

// Class for one dimension quadrics for example: Sphere and Teapot

class OneDimension : public Quadric
{
public:	
	void ChangeDimensions(float Radius);
	void DrawBox();
};


// -------------
// Cylinder
// -------------

class Cylinder: public TwoDimension
{
public:
	Cylinder(GLuint Name);

	void Draw();
	void Save(wxString& XMLContent);
};

// -------------
// Sphere
// -------------

class Sphere : public OneDimension
{
public:
	Sphere(GLuint Name);
	
	void Draw();
	void Save(wxString& XMLContent);
};

// -------------
// Cone
// -------------

class Cone: public TwoDimension
{
public:
	Cone(GLuint Name);
	void Draw();
	void Save(wxString& XMLContent);
};

// -------------
// Teapot
// -------------

class Teapot : public OneDimension
{
public:
	Teapot(GLuint Name);
	void Draw();
	void Save(wxString& XMLContent);
};

// -------------
// Tetrahedron
// -------------

class Tetrahedron : public OneDimension
{
public:
	Tetrahedron(GLuint Name);
	void Draw();
	void Save(wxString& XMLContent);
};

// -------------
// Octahedron
// -------------

class Octahedron : public OneDimension
{
public:
	Octahedron(GLuint Name);
	void Draw();
	void Save(wxString& XMLContent);
};

// -------------
// Dodecahedron
// -------------

class Dodecahedron : public OneDimension
{
public:
	Dodecahedron(GLuint Name);
	void Draw();
	void Save(wxString& XMLContent);
};

// -------------
// Icosahedron
// -------------

class Icosahedron : public OneDimension
{
public:
	Icosahedron(GLuint Name);
	void Draw();
	void Save(wxString& XMLContent);
};

// -------------------
// Rombic Dodecahedron
// -------------------

class RombicDodecahedron : public OneDimension
{
public:
	RombicDodecahedron(GLuint Name);
	void Draw();
	void Save(wxString& XMLContent);
};

// -------------
// Torus
// -------------

class Torus : public TwoDimension
{
public:
	Torus(GLuint Name);
	void Draw();
	void DrawBox();
	void Save(wxString& XMLContent);
	void Load(xml_node<>* Node);
	void ChangeDimensions(float radius, float height);
};

// -------------
// Text 3D
// -------------

class Text : public TwoDimension
{
public:	
	Text(GLuint Name);
	void Draw();
	void DrawBox();
	void Save(wxString& XMLContent);
	void ChangeDimensions(float radius, float height);
	void Load(xml_node<>* Node);
	void SetText(wxString Str);
	void Copy(Text* t);
	void DoCopy(wxString& Str);
private:
	wxString m_Str;
};

#endif _QUADRIC_H_