/*****************************************************************************
   ISSIGraph: Experimental Surface Editor Prototype						 
   Copyright (C) 2013 by Carlos Jimnez de Parga								 
   
   Trabajo para el:
   MASTER EN INVESTIGACIN EN INGENIERA DE SOFTWARE Y SISTEMAS INFORMTICOS
			REPRESENTACIN GRFICA DE SUPERFICIES IMPLCITAS			
	DEPARTAMENTO DE INGENIERA DE SOFTWARE Y SISTEMAS INFORMTICOS 
							- UNED -

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
   *************************************************************************/

/* 
	FILE: Surface.h
	PURPOSE: NURBS and Bezier surfaces main class (header)
	Author: Carlos Jimnez de Parga (c)
	License: GPL v3.0
*/

#ifndef _SURFACE_H_
#define _SURFACE_H_

#include "box.h"
#include "tool.h"

// ----------------------------------------------------------------------------
// Control Point class
// ----------------------------------------------------------------------------

class CtrlPoint : public VertexBox
{
public:
	CtrlPoint(GLuint name); // Default constructor
	CtrlPoint(GLuint name, float ptx, float pty, float ptz, float angleX, float angleY, float width, float height);
	void Draw();
	void DrawBox();
	Tool::Vector3D GetVectorT();
	Tool::Vector3D GetVectorM();
	void SetVector(Tool::Vector3D Vector);
};

// Declare list of control points

WX_DECLARE_LIST(CtrlPoint, ControlPoint);

// ----------------------------------------------------------------------------
// Surface class: base class for Nurbs and Bezier
// ----------------------------------------------------------------------------


class Surface : public VertexBox
{
public:
	Surface(int Limit);
	~Surface();
	void Move(int x, int y, float SceneX, float SceneY, int screenWidth, int screenHeight);
	virtual void DrawSelection()=0;
	virtual bool GetReady()=0;
	void ResetPass();
	void SetCtrlPoints(ControlPoint* CtrlPts);
	ControlPoint* GetCtrlPoints();
	void MakeAverage();	
	void MakeAllAverage();
	void Copy(Surface* s);
	void ProcessCtrlPoints();
	bool CheckLimit();
	void Increment();
	void Decrement();
	bool Selected();
	void Spin(float xSpin, float ySpin);
	void Destroy();
	void Save(wxString& XMLContent);
	unsigned int CtrlCPLoad(xml_node<>* Node);
	void SetMkAverage(bool Avg);
	enum StateRot {ID_FROM_GENERALROTATE, ID_FROM_LOCALROTATE, ID_ROT_NOTHING};
	StateRot GetRotState();
	void SetRotState(StateRot State);
protected:
	void DoCopy(int LimitCount, float OldTX, float OldTY, float tx, float ty, float MaxX, float MaxY, float MinX, float MinY, ControlPoint* cp);
	
	bool m_MkAvg;
	int m_Limit, m_LimitCount;
	float m_OldTX, m_OldTY;
	float m_MaxX,m_MaxY,m_MinX,m_MinY;
	ControlPoint* m_ControlPointList;
	GLfloat m_CtrlPoint[8][8][3];
	bool m_Pass;
	Tool::Vector3D m_Average;
	StateRot m_RotState;
};

// ----------------------------------------------------------------------------
// Nurbs class
// ----------------------------------------------------------------------------

class NURBS : public Surface
{
public:
	NURBS(int Limit, GLuint name);
	~NURBS();

	void DrawBox();
	void Draw();
	void DrawSelection();
	void Destroy();
	bool GetReady();
	void RecalculateDeletedCP();
	void Copy(NURBS* Nurbs);
	void Save(wxString& XMLContent);
	void Load(xml_node<>* Node);
protected:
	void DoCopy(int UpToRender);

private:
	void GenerateKnots();
	GLUnurbsObj *m_Nurb;
	GLfloat* m_KnotsU;
	GLfloat m_KnotsV[16];
	unsigned int m_UpToRender;
	int m_OldCount;
};

// ----------------------------------------------------------------------------
// Bezier class
// ----------------------------------------------------------------------------

class BEZIER : public Surface
{
public:
	BEZIER(int Limit, GLuint name);

	void DrawBox();
	void Draw();
	void DrawSelection();
	bool GetReady();
	void Save(wxString& XMLContent);
};

#endif _SURFACE_H_