/* 
	FILE: World.cpp
	PURPOSE: World and render bussines logic (implementation)
	Author: Carlos Jimnez de Parga (c)
	License: GPL v3.0
*/

#include "world.h"
#include "tool.h"
#include "wx/listimpl.cpp"



WX_DEFINE_LIST(ControlPoint);
WX_DEFINE_LIST(ListBox);

// ----------------------------------------------------------------------------
// Axis class implementation
// ----------------------------------------------------------------------------

Axis::Axis(GLuint name,float ptx,float pty, float ptz, float width, float height):VertexBox(ptx,pty,ptz,width,height)
{
	m_Name = name;
}

void Axis::Spin(float xSpin, float ySpin)
{
	m_xAngle+=xSpin;
	m_yAngle+=ySpin;

	m_xAngle %= 360;
	m_yAngle %= 360;
}

// Draws the axis

void Axis::Draw()
{
	
	glPushMatrix();
	
	glTranslatef(m_tx ,m_ty,m_tz);
	glColor4f(0.5f,0.5f,0.5f,0.0f);
	glLoadName(m_Name);
	glBegin(GL_QUADS);
		glVertex3f(- 2.5f,2.5f,0.0f);
		glVertex3f(2.5f,2.5f,0.0f);
		glVertex3f(2.5f,-2.5f,0.0f);
		glVertex3f(-2.5f,-2.5f,0.0f);
	glEnd();

	glRotatef(m_xAngle, 1.0f, 0.0f, 0.0f);
    glRotatef(m_yAngle, 0.0f, 1.0f, 0.0f);
	glBegin(GL_LINES);
        glColor3f(1.0f,0.0f,0.0f);
		glVertex3f(0.0f, 0.0f, 0.0f);
        glVertex3f(2.0f, 0.0f, 0.0f);
		glColor3f(0.0f,1.0f,0.0f);
		glVertex3f(0.0f, 0.0f, 0.0f);
        glVertex3f(0.0f, 2.0f, 0.0f);
		glColor3f(0.0f,0.0f,1.0f);
		glVertex3f(0.0f, 0.0f, 0.0f);
        glVertex3f(0.0f, 0.0f, 2.0f);
	glEnd();

	glPushMatrix();
	glTranslatef(2.0f,0.0f,0.0f);
	glColor3f(1.0f,0.0f,0.0f);
	glScalef(0.007f, 0.007f, 0.007f);
	Textout::Print("X");
	glPopMatrix();
	
	glPushMatrix();
	glTranslatef(0.0,2.0f,0.0f);
	glColor3f(0.0f,1.0f,0.0f);
	glScalef(0.007f, 0.007f, 0.007f);
	Textout::Print("Y");
	glPopMatrix();
	
	glPushMatrix();
	glTranslatef(0.0f,0.0f,2.0f);
	glColor3f(0.0f,0.0f,1.0f);
	glScalef(0.007f, 0.007f, 0.007f);
	Textout::Print("Z");
	glPopMatrix();

	glPopMatrix();

}

// Accessors

float Axis::GetAngleX()
{
	return m_xAngle;
}

float Axis::GetAngleY()
{
	return m_yAngle;
}

void Axis::SetAngleX(float Xangle)
{
	m_xAngle = Xangle;
}

void Axis::SetAngleY(float Yangle)
{
	m_yAngle = Yangle;
}


// ----------------------------------------------------------------------------
// Memory class implementation: Performs Copy/Cut/Paste
// ----------------------------------------------------------------------------

Memory::Memory()
{
	m_Mem = NULL;
	
	m_MouseX = 0;
	m_MouseY = 0;
	m_UndoX = 0;
	m_UndoY = 0;
}

Memory::~Memory()
{
	if (m_Mem != NULL)
		delete m_Mem;
		
}

// Performs copy/cut

void Memory::OnCopy(Box* selected, GLuint count)
{
	if (m_Mem != NULL)
		delete m_Mem;
		
	m_Mem = selected;
	m_Mem = switchClass(count);

}

// Performs paste

Box* Memory::OnPaste(int x, int y, float SceneX, float SceneY, int scrW, int  scrH, GLuint count)
{
	Box* memAux;
	if (m_Mem)
	{
		memAux = switchClass(count);
		Tool::Vector3D vPaste = Tool::Math::GetOGLPos(x,y,scrW,scrH);
		
		/*wxMessageBox(Tool::Math::ConvertNumber(memAux->GetWidth()));
		wxMessageBox(Tool::Math::ConvertNumber(memAux->GetHeight()));
		wxMessageBox(Tool::Math::ConvertNumber(vPaste.m_x));
		wxMessageBox(Tool::Math::ConvertNumber(vPaste.m_y));*/

		if (dynamic_cast<Text*>(memAux)) 
			memAux->Paste(vPaste.m_x, vPaste.m_y,SceneX, SceneY);
		else if (dynamic_cast<Surface*>(memAux))
			memAux->Move(x,y,SceneX,SceneY,scrW,scrH);
		else memAux->Paste(vPaste.m_x + memAux->GetWidth(), vPaste.m_y - memAux->GetHeight(),SceneX, SceneY);
		return memAux;
	}
	return NULL;
}

// Select type of object

Box* Memory::switchClass(GLuint Count)
{
	
	if (Cone* OldCon = dynamic_cast<Cone*>(m_Mem))
	{
		Cone* NewCon = new Cone(0);
		OldCon->Copy(NewCon);
		NewCon->SetName(Count);
		return NewCon;
	}
	else if (Cylinder* OldCyl = dynamic_cast<Cylinder*>(m_Mem))
	{
		Cylinder* NewCyl = new Cylinder(0);
		OldCyl->Copy(NewCyl);
		NewCyl->SetName(Count);
		return NewCyl;
	}
	else if (Sphere* OldSph = dynamic_cast<Sphere*>(m_Mem))
	{
		Sphere* NewSph = new Sphere(0);
		OldSph->Copy(NewSph);
		NewSph->SetName(Count);
		return NewSph;
	} 
	else if (Teapot* OldTpt = dynamic_cast<Teapot*>(m_Mem))
	{
		Teapot* NewTpt = new Teapot(0);
		OldTpt->Copy(NewTpt);
		NewTpt->SetName(Count);
		return NewTpt;
	}
	else if (Tetrahedron* OldTh = dynamic_cast<Tetrahedron*>(m_Mem))
	{
		Tetrahedron* NewTh = new Tetrahedron(0);
		OldTh->Copy(NewTh);
		NewTh->SetName(Count);
		return NewTh;
	}
	else if (Octahedron* OldOh = dynamic_cast<Octahedron*>(m_Mem))
	{
		Octahedron* NewOh = new Octahedron(0);
		OldOh->Copy(NewOh);
		NewOh->SetName(Count);
		return NewOh;
	}
	else if (Dodecahedron* OldDh = dynamic_cast<Dodecahedron*>(m_Mem))
	{
		Dodecahedron* NewDh = new Dodecahedron(0);
		OldDh->Copy(NewDh);
		NewDh->SetName(Count);
		return NewDh;
	}
	else if (Icosahedron* OldIh = dynamic_cast<Icosahedron*>(m_Mem))
	{
		Icosahedron* NewIh = new Icosahedron(0);
		OldIh->Copy(NewIh);
		NewIh->SetName(Count);
		return NewIh;
	}
	else if (RombicDodecahedron* OldRdh = dynamic_cast<RombicDodecahedron*>(m_Mem))
	{
		RombicDodecahedron* NewRdh = new RombicDodecahedron(0);
		OldRdh->Copy(NewRdh);
		NewRdh->SetName(Count);
		return NewRdh;
	}
	else if (Torus* OldTor = dynamic_cast<Torus*>(m_Mem))
	{
		Torus* NewTor = new Torus(0);
		OldTor->Copy(NewTor);
		NewTor->SetName(Count);
		return NewTor;
	}
	else if (Text* OldTxt = dynamic_cast<Text*>(m_Mem))
	{
		Text* NewTxt = new Text(0);
		OldTxt->Copy(NewTxt);
		NewTxt->SetName(Count);
		return NewTxt;
	}
	else if (NURBS* OldNrb = dynamic_cast<NURBS*>(m_Mem))
	{
		NURBS* NewNrb = new NURBS(MAX_CTRL_POINTS,0);
		NewNrb->SetCtrlPoints(new ControlPoint());
		OldNrb->Copy(NewNrb);
		NewNrb->SetName(Count);
		return NewNrb;
	}  else if (BEZIER* OldBzr = dynamic_cast<BEZIER*>(m_Mem))
	{
		BEZIER* NewBzr = new BEZIER(MAX_CTRL_POINTS,0);
		NewBzr->SetCtrlPoints(new ControlPoint());
		OldBzr->Copy(NewBzr);
		NewBzr->SetName(Count);
		return NewBzr;
	} else return NULL;
}

void Memory::SetMouse(int x, int y)
{
	m_MouseX = x;
	m_MouseY = y;
}

Tool::COORDS Memory::GetMouse()
{
	Tool::COORDS Cords(m_MouseX,m_MouseY);
	return Cords;
}

void Memory::SetUndo(Tool::COORDS Cords)
{
	m_UndoX = Cords.m_x;
	m_UndoY = Cords.m_y;
}


Tool::COORDS Memory::GetUndo()
{
	Tool::COORDS Cords(m_UndoX, m_UndoY);
	return Cords;
}


// ----------------------------------------------------------------------------
// World class implementation: Performs scene container
// ----------------------------------------------------------------------------

World::World(wxGLCanvas *Canvas, wxFrame* Frame)
{
	m_Canvas = Canvas;
	m_Frame = Frame;
	Create();
}

void World::Create()
{
	m_ListBox = new ListBox();
	m_ListBox->Append(m_CordSys = new Axis(99999999,-20.0f,-20.0f,0.0f,2.5f,2.5f));
	m_ButtonLDown = m_ButtonRDown = m_ZoomPlus = false;
	m_SceneX = m_SceneY = m_DepthZ = 0.0f;
	m_RotateX = m_RotateY = 0;
	m_CountObject = 1;
	m_Saving = m_InModeling = m_InModelMove = m_InAction = m_InCopy = m_ComeFromLocal = m_ComeFromClick = false;
	m_SelectedBox = m_SelectedCP = NULL;
	m_FromUndo = ID_FROM_NOTHING;
}


// Renders the scene

void World::Render(bool Lactive)
{
	
	glPushMatrix();

	glTranslatef(-m_SceneX,-m_SceneY,m_DepthZ); // Translate all the scene according to depth

	if (m_SelectedCP && !m_Saving) m_SelectedCP->DrawBox();
	

	// Iterates through all objects in scene
	
	for(ListBox::iterator it = m_ListBox->begin(); it!=m_ListBox->end();it++)
	{
		Surface* s = dynamic_cast<Surface*>(*it);
		
		if (s) // If surface class
		{
						
			s->ProcessCtrlPoints();
			
	
			if (s->GetReady() && !Lactive) // Lactive (if comes from Main->picking)
				s->Draw();
						
			if (!m_Saving) // The user is not exporting to image
				for(ControlPoint::iterator at = s->GetCtrlPoints()->begin();at != s->GetCtrlPoints()->end();at++){
					CtrlPoint *cp = *at;
					cp->Draw(); // Draws all 3D objects
				}		
				
			
			if (m_SelectedBox == (*it) && !m_Saving) 
				s->DrawBox();	// Draws selection box if not exporting


			if (!m_Saving)	
				s->DrawSelection(); // Draws selection handlers
			
			
		}
		else // Another 3D object
		{		

				(*it)->Draw();

				if (m_SelectedBox == (*it) && !m_Saving) 
					(*it)->DrawBox();
			
				Quadric* quad = dynamic_cast<Quadric*>(*it);
				if (quad != NULL && !m_Saving)
					quad->DrawSelection();
		
		}	
	}


	glPopMatrix();
}

// When the user presses the left button of the mouse

void World::OnMouseLDown(Tool::Idents* Idents)
{
		
	m_ButtonLDown = true;


	float SelectedAngleX =0.0f, SelectedAngleY = 0.0f;

	
	if (m_SelectedBox)
	{
		SelectedAngleX = m_SelectedBox->GetAngleX();
		SelectedAngleY = m_SelectedBox->GetAngleY();
	}

	m_Frame->SetStatusText(wxString::Format(_("X-angle = %.3f :: Y-angle = %.3f"), SelectedAngleX, SelectedAngleY),2);
	m_Frame->SetStatusText(_("Ready to get orders"),4);

	
	// Detects if any 3D object is under the mouse coordinates

	
	for(ListBox::iterator it = m_ListBox->begin();it != m_ListBox->end();it++)
	{
		Box *box = *it;
		for(int i=0;i < Idents->GetHits(); i++)
		{
			if (box->CompareName(Idents->GetItem(i))) // Identify the 3D object in OGL scene
			{
				m_SelectedBox = box;
				m_UndoCords.m_x = m_SelectedBox->GetCoords().m_x;
				m_UndoCords.m_y = m_SelectedBox->GetCoords().m_y;	
		
				m_CordSys->SetAngleX(m_SelectedBox->GetAngleX()); // Informs the coordinate system
				m_CordSys->SetAngleY(m_SelectedBox->GetAngleY());
				Surface* s = dynamic_cast<Surface*>(m_SelectedBox);
				if(s != NULL) {
					s->ResetPass();
					s->SetMkAverage(true);
					
				} 
				
				m_SelectedCP = NULL;
				m_ComeFromClick = true;

				delete Idents; // Delete the idents array object
				
				m_InCopy = m_InAction = false;
							
				return;
			} 
		}
	}

	// Check if a CP is picked

	Surface* s = dynamic_cast<Surface*>(m_SelectedBox);
	if(s != NULL) {
			for(ControlPoint::iterator it = s->GetCtrlPoints()->begin();it !=s->GetCtrlPoints()->end();it++)
			{
				CtrlPoint *cp = *it;
		
				for(int i=0;i < Idents->GetHits(); i++) // Indetify the Control Point
					if (cp->CompareName(Idents->GetItem(i)))
					{
						m_SelectedCP = cp; // Select it
						
						m_UndoCords.m_x = m_SelectedCP->GetCoords().m_x;
						m_UndoCords.m_y = m_SelectedCP->GetCoords().m_y;

						delete Idents;
						
						m_InAction = false;
						m_Frame->SetStatusText(_("Ready to get orders"),4);

						return;
					}
			}
	
	} 
	
	m_SelectedBox = NULL; // Nothing selected
	m_SelectedCP = NULL;

	delete Idents;
}

// Used when the user presses the right button of the mouse

void World::OnMouseRDown(int x, int y, Tool::Idents* Idents, int ScreenWidth, int ScreenHeight)
{
	const float ogx = Tool::Math::GetOGLPos(x,y,ScreenWidth,ScreenHeight).m_x + m_SceneX;
	const float ogy = Tool::Math::GetOGLPos(x,y,ScreenWidth,ScreenHeight).m_y + m_SceneY;
	
	m_ButtonRDown = true;
	m_InModeling = false;

	if (Surface* Surf = dynamic_cast<Surface*>(m_SelectedBox))
	{
		for(int i = 0;i < Idents->GetHits(); i++)
		{	
			if (Surf->CompareName(Idents->GetItem(i))) // Identify if we are over surface selector
			{
				delete Idents;
				return;
			}
		}
	}

	if (m_SelectedCP)
	{
			for(int i=0;i < Idents->GetHits(); i++) // Indetify if over the selected control point
				if (m_SelectedCP->CompareName(Idents->GetItem(i)))
				{
					delete Idents;
					return;
				}
	}

	delete Idents;
	

	if (ogx < MAX_X && ogx > MIN_X && ogy < MAX_Y && ogy > MIN_Y)
	{	
			Surface* v = dynamic_cast<Surface*>(m_SelectedBox);
			if (v != NULL && !v->CheckLimit() && !m_InCopy) // Add control point to the scene
			{
				v->GetCtrlPoints()->Append(new CtrlPoint(m_CountObject++,ogx, ogy, 1.0f,0.0f,0.0f,0.5f,0.5f));
				v->Increment();
				
				if (!m_InModeling)
				{
					m_Frame->SetStatusText(_("Modeling..."),4);
					m_InAction = false;
					m_InModeling = true;
				}

				
			} else if (v != NULL && v->CheckLimit()) 
			{
				m_InModeling = false;
				::wxSetCursor(wxCursor(wxCURSOR_ARROW));
			} 
				
	}		

}
				

// Used when the user moves the mouse

void World::OnMouseMove(int x, int y, int ScreenWidth, int ScreenHeight)
{	
	TwoDimension* Two;
	OneDimension* One;
	
	m_MouseX = x;
	m_MouseY = y;

	m_ScreenWidth = ScreenWidth;
	m_ScreenHeight = ScreenHeight;


	if (m_ButtonLDown && m_SelectedCP) // Move if selected control point
	{
		m_Frame->SetStatusText(_("Action moving control point"),4);
		m_SelectedCP->Move(x,y, m_SceneX, m_SceneY, ScreenWidth, ScreenHeight);
		m_SelectedCP->SetPassV(true);
		if (Surface* s = dynamic_cast<Surface*>(m_SelectedBox))
			s->SetMkAverage(true);
		::wxSetCursor(wxCursor(wxCURSOR_HAND));
	}
	else if (m_ButtonLDown && m_SelectedBox) //Move if selected box 3D object
	{
		m_Frame->SetStatusText(_("Action moving surface"),4);
		m_SelectedBox->Move(x, y, m_SceneX, m_SceneY, ScreenWidth, ScreenHeight);

		if (Surface* s = dynamic_cast<Surface*>(m_SelectedBox))
			s->SetMkAverage(true);
		::wxSetCursor(wxCursor(wxCURSOR_HAND));
	} else if (m_InModelMove && (Two = dynamic_cast<TwoDimension*>(m_SelectedBox)))
		ChangeTwoDim(x, y, Two);
	  else if (m_InModelMove && (One = dynamic_cast<OneDimension*>(m_SelectedBox)))
		ChangeOneDim(x, One);

	
	float SelectedX = 0.0f;
	float SelectedY = 0.0f;
	float SelectedZ = 0.0f;

	if (m_SelectedBox)
	{
		SelectedX = m_SelectedBox->GetCoords().m_x;
		SelectedY = m_SelectedBox->GetCoords().m_y;
		SelectedZ = m_SelectedBox->GetCoords().m_z;
	}
	
	m_Frame->SetStatusText(wxString::Format(_("X = %.3f :: Y = %.3f :: Z = %.3f"), SelectedX, SelectedY, SelectedZ),1);
	m_Frame->SetStatusText(wxString::Format(_("MouseX = %3i :: MouseY = %3i"), x, y), 3);
	if (!(m_InModelMove || m_InModeling || m_InAction || m_ButtonLDown) )
		m_Frame->SetStatusText(_("Action moving"),4);
}

// Used when the user raises the left button

void World::OnMouseLUp(int x, int y)
{
	m_ButtonLDown = false;

	if (m_SelectedCP)
		m_SelectedCP->SetMouse(x, y);
	else if (m_SelectedBox) m_SelectedBox->SetMouse(x,y);
	
	if ((m_SelectedCP) && ((m_SelectedCP->GetCoords().m_x != m_UndoCords.m_x) && (m_SelectedCP->GetCoords().m_y != m_UndoCords.m_y)))
	{
		m_SelectedCP->m_isMoved = true;
		m_SelectedCP->SetUndoToggle(true);	
		m_SelectedCP->SetUndo(m_UndoCords);
	}

	if ((m_SelectedBox) && ((m_SelectedBox->GetCoords().m_x != m_UndoCords.m_x) && (m_SelectedBox->GetCoords().m_y != m_UndoCords.m_y)))
	{
		m_SelectedBox->m_isMoved = true;
		m_SelectedBox->SetUndoToggle(true);	
		m_SelectedBox->SetUndo(m_UndoCords);
	} 
}

// Used when the user raises the right button

void World::OnMouseRUp()
{
	m_ButtonRDown = false;
	if (m_InModelMove)
		m_InModelMove = false;
}

// Used when the user moves the mouse wheel

void World::OnMouseWheel(int zooming)
{
	if (zooming > 0) 
	{
		if (m_SelectedCP) m_SelectedCP->ZoomMinus();
		else if (m_SelectedBox) m_SelectedBox->ZoomMinus();
		m_Frame->SetStatusText(_("Action zooming Z-"),4);
	} else {
		if (m_SelectedCP) m_SelectedCP->ZoomPlus();
			else if (m_SelectedBox) m_SelectedBox->ZoomPlus();
		
		m_Frame->SetStatusText(_("Action zooming Z+"),4);
	
	}
	
}


// Changes Cylinder and Cone dimensions

void World::ChangeTwoDim(int x, int y, TwoDimension* Two)
{
	if ((x > 1.0f) && (y > 1.0f))
		Two->ChangeDimensions(x/20.0f,y/20.0f);
}

// Changes Sphere dimensions

void World::ChangeOneDim(int x, OneDimension* One)
{
	if (x > 1.0f) 
		One->ChangeDimensions(x/20.0f);
}

// Spin the object around its own axis

void World::OnSpin(float XSpin, float YSpin)
{
	Surface* s;
	ControlPoint::iterator at;
	
	s = dynamic_cast<Surface*>(m_SelectedBox);
	if (s) // If surface spin all control points
	{	
		s->Spin(XSpin, YSpin);
		
		if (s->GetRotState() == Surface::ID_FROM_GENERALROTATE || s->GetRotState() == Surface::ID_ROT_NOTHING)
		{
			for(ControlPoint::iterator xt = s->GetCtrlPoints()->begin();xt !=s->GetCtrlPoints()->end();xt++)
			{	
				(*xt)->SetPassV(true);
			}
	
			s->SetRotState(Surface::ID_FROM_LOCALROTATE); // Change state (we come from spin object)
			s->SetMkAverage(true);
		}

		s->MakeAllAverage();
		s->SetMkAverage(false);

		for(at = s->GetCtrlPoints()->begin();at != s->GetCtrlPoints()->end();at++){
				CtrlPoint *cp = *at;
				cp->Spin(XSpin,YSpin);
			}
	} else	m_SelectedBox->Spin(XSpin, YSpin);

	m_CordSys->Spin(XSpin, YSpin);
			
	m_ComeFromLocal = true;
}

// Rotate the whole scene

void World::RotateScene()
{

	if (m_ComeFromLocal || m_ComeFromClick)
	{
		m_CordSys->SetAngleX(m_RotateX); // Informs the scene coordinate system
		m_CordSys->SetAngleY(m_RotateY);
		m_ComeFromLocal = false;
		m_ComeFromClick = false;
	}

	for(ListBox::iterator it = m_ListBox->begin(); it!=m_ListBox->end();it++)
	{
		
		if (Surface* s = dynamic_cast<Surface*>((*it))) // Swap rotations
		{
			if (s->GetRotState() == Surface::ID_FROM_LOCALROTATE || s->GetRotState() == Surface::ID_ROT_NOTHING)
			{
				for(ControlPoint::iterator xt = s->GetCtrlPoints()->begin();xt !=s->GetCtrlPoints()->end();xt++)
				{
					(*xt)->SetPassR(true);
				}
			
				s->SetRotState(Surface::ID_FROM_GENERALROTATE); // Change state (we come from rotate scene)
			}
		
			// Rotates control points
			for(ControlPoint::iterator xt = s->GetCtrlPoints()->begin();xt !=s->GetCtrlPoints()->end();xt++)
			{
				(*xt)->RotateScene(m_RotateX, m_RotateY);
			}
		 } else (*it)->RotateScene(m_RotateX, m_RotateY); // Rotates another object
		
	}
				
}

// KeyCode processing

bool World::OnKeyDown(int keyCode)
{	
    switch (keyCode)
    {
        case WXK_RIGHT: // Y+ spin
           	if (m_SelectedBox) 
			{
				OnSpin(0.0, ANGLE_SPIN);
				m_Frame->SetStatusText(wxString::Format(_("X-angle = %3.f, Y-angle = %.3f"), m_CordSys->GetAngleX(), m_CordSys->GetAngleY()),2);
				if (!(m_InModelMove || m_InAction || m_ButtonLDown) )
				m_Frame->SetStatusText(_("Action rotating Y+"),4);
			}
			break;

        case WXK_LEFT: // Y- spin
           		if (m_SelectedBox)
				{	
					OnSpin(0.0, -ANGLE_SPIN);
					m_Frame->SetStatusText(wxString::Format(_("X-angle = %.3f, Y-angle = %.3f"), m_CordSys->GetAngleX(), m_CordSys->GetAngleY()),2);
					if (!(m_InModelMove || m_InAction || m_ButtonLDown) )
							m_Frame->SetStatusText(_("Action rotating Y-"),4);
				}
			break;

        case WXK_DOWN: // X- spin
            	if (m_SelectedBox)
				{	
					OnSpin(-ANGLE_SPIN, 0.0);
					m_Frame->SetStatusText(wxString::Format(_("X-angle = %.3f, Y-angle = %.3f"), m_CordSys->GetAngleX(), m_CordSys->GetAngleY()),2);
					if (!(m_InModelMove || m_InAction || m_ButtonLDown) )
							m_Frame->SetStatusText(_("Action rotating X-"),4);
				}
				break;

        case WXK_UP: // X- spin
				if (m_SelectedBox)
				{	
					OnSpin(ANGLE_SPIN, 0.0);
					m_Frame->SetStatusText(wxString::Format(_("X-angle = %.3f :: Y-angle = %.3f"), m_CordSys->GetAngleX(), m_CordSys->GetAngleY()),2);
					if (!(m_InModelMove || m_InAction || m_ButtonLDown) )
							m_Frame->SetStatusText(_("Action rotating X+"),4);
				}
			break;
		case WXK_NUMPAD_SUBTRACT: // Changes zoom
			if (m_DepthZ > MIN_Z) m_DepthZ-=2;
			m_Frame->SetStatusText(_("Zooming Z-"),4);
			break;
		case WXK_NUMPAD_ADD:
			if (m_DepthZ < MAX_Z) m_DepthZ+=2;
			m_Frame->SetStatusText(_("Zooming Z+"),4);
			break;
		case WXK_NUMPAD8:
			m_RotateX+=ANGLE_SCENE;
			m_RotateX %= 360;
			RotateScene();
			m_CordSys->Spin(ANGLE_SCENE,0.0f);
			break;
	    case WXK_NUMPAD2:
			m_RotateX-=ANGLE_SCENE;
			m_RotateX %= 360;
			RotateScene();
			m_CordSys->Spin(-ANGLE_SCENE,0.0f);
			break;
		case WXK_NUMPAD4:
			m_RotateY-=ANGLE_SCENE;
			m_RotateY %= 360;
			RotateScene();
			m_CordSys->Spin(0.0f, -ANGLE_SCENE);
			break;
	    case WXK_NUMPAD6:
			m_RotateY+=ANGLE_SCENE;
			m_RotateY %= 360;
			RotateScene();
			m_CordSys->Spin(0.0f, ANGLE_SCENE);
			break;  
		case WXK_DELETE: // Delete the selected object
			DeleteSelected();
			m_Frame->SetStatusText(_("Object deleted"),4);
			m_InAction = true;
			break;
		case WXK_SPACE:
			break;
		default:
			return true;
    }
	return false;
}

// When the user selects a button in 3D IssiControl dialog

void World::OnSetControlOrder(int OrderId)
{
	switch(OrderId)
	{
		case wxID_ZOOMMINUS:
			if (m_DepthZ > MIN_Z) m_DepthZ-=2;
				m_Frame->SetStatusText(_("Zooming Z-"),4);
			break;
		case wxID_ZOOMPLUS:
			if (m_DepthZ < MAX_Z) m_DepthZ+=2;
			m_Frame->SetStatusText(_("Zooming Z+"),4);
			break;
		case wxID_ROTATEYPLUS:
			if (m_SelectedBox)
			{
				OnSpin(0.0, ANGLE_SPIN);
				m_Frame->SetStatusText(wxString::Format(_("X-angle = %3.f, Y-angle = %.3f"), m_CordSys->GetAngleX(), m_CordSys->GetAngleY()),2);
				if (!(m_InModelMove || m_InAction || m_ButtonLDown) )
				m_Frame->SetStatusText(_("Action rotating Y+"),4);
			} else wxMessageBox("You must select a 3D object");
			break;
		case wxID_ROTATEYMINUS:
			if (m_SelectedBox)
			{	
				OnSpin(0.0, -ANGLE_SPIN);
				m_Frame->SetStatusText(wxString::Format(_("X-angle = %.3f, Y-angle = %.3f"), m_CordSys->GetAngleX(), m_CordSys->GetAngleY()),2);
				if (!(m_InModelMove || m_InAction || m_ButtonLDown) )
				m_Frame->SetStatusText(_("Action rotating Y-"),4);
			} else wxMessageBox("You must select a 3D object");
			break;
		case wxID_LEFT:
			m_SceneX-=3.0f;
			break;
		case wxID_RIGHT:
			m_SceneX+=3.0f;
			break;
		case wxID_UP:
			m_SceneY+=3.0f;
			break;
		case wxID_DOWN:
			m_SceneY-=3.0f;
			break;
		case wxID_CENTER:
			m_SceneX = 0.0f;
			m_SceneY = 0.0f;
			break;
	}
}

// Delete the selected 3D object

void World::DeleteSelected()
{
	bool deleted = false;
	ControlPoint::iterator itAuxCP;
	ListBox::iterator itAuxLB;
	
	Surface* s = dynamic_cast<Surface*>(m_SelectedBox);
	if(s != NULL && m_SelectedCP) {
			for(ControlPoint::iterator it = s->GetCtrlPoints()->begin();it !=s->GetCtrlPoints()->end();it++)
			{
				CtrlPoint* cp = *it;
				if (m_SelectedCP == cp)
				{
					itAuxCP = it;
					delete cp;
					deleted = true;
					break;
				}
			}			
		
		if (deleted)
		{
			m_SelectedCP = NULL;
			s->GetCtrlPoints()->erase(itAuxCP);
			s->Decrement();
			NURBS* Nurbs = dynamic_cast<NURBS*>(s);
			if (Nurbs)
				Nurbs->RecalculateDeletedCP();
		}
	}
	if (m_SelectedBox && !deleted)
	{
			deleted = false;

			for(ListBox::iterator it = m_ListBox->begin(); it!=m_ListBox->end();it++)
			{
				Box* box  = *it; // Search it on the list. This O(n) is "ok" enough.
				if (box == m_SelectedBox && box!=m_CordSys)
				{
					if (dynamic_cast<Surface*>(m_SelectedBox))
						m_InModeling = false;

					itAuxLB = it;
					delete box;
					deleted = true;
					break;
				}
			}

			if (deleted) // Delete it
			{
				
				m_SelectedBox = m_SelectedCP = NULL;
				m_ListBox->erase(itAuxLB);
			
			}
	}
}

// Handles Copy/Cut/Paste

void World::OnCopyPaste(int Id)
{
	Box* Pasted;

	switch(Id)
	{
		case ID_COPY:
			if (m_SelectedBox && !m_SelectedCP)
			{
				m_FromUndo = ID_FROM_COPY;
				m_Mem.OnCopy(m_SelectedBox, m_CountObject++);
				m_InCopy = true;
			}
			break;

		case ID_CUT:
			if (m_SelectedBox && !m_SelectedCP)
			{
				Tool::COORDS Cords(m_SelectedBox->GetMouse().m_x, m_SelectedBox->GetMouse().m_y);
				
				m_Mem.SetUndo(Cords);
			
				m_Mem.OnCopy(m_SelectedBox, m_CountObject++);	
				DeleteSelected();
				m_FromUndo = ID_FROM_CUT;
				m_InCopy = true;
			}
			break;

		case ID_PASTE:
			m_SelectedBox = Pasted = m_Mem.OnPaste(m_MouseX, m_MouseY, m_SceneX, m_SceneY, m_ScreenWidth, m_ScreenHeight, m_CountObject++);
			if (Pasted) // Add to the 3D object list
			{	
				m_ListBox->Append(Pasted);
				m_InCopy = false;

				Surface* Surf = dynamic_cast<Surface*>(m_SelectedBox);
				if (Surf)	
					m_CountObject += Surf->GetCtrlPoints()->GetCount();
			}
			break;

	}
}

// Accessors

bool World::GetInModeling()
{
	return m_InModeling;
}

bool World::GetInModelMove()
{
	return m_InModelMove;
}

void World::OnSaving(bool Saving)
{
	m_Saving = Saving; 
	Render(false);
}

void World::OnDoUndoRedo()
{
	if (m_SelectedCP) // Move if selected control point
	{
			Tool::COORDS CurCords,OldCords;

			if (m_SelectedCP->isUndoToggle()) // Toggle UNDO / REDO
			{
				OldCords = m_SelectedCP->GetUndo();
				
				if (m_SelectedCP->m_isMoved) // A new position happens
				{
					CurCords.m_x = m_SelectedCP->GetCoords().m_x;
					CurCords.m_y = m_SelectedCP->GetCoords().m_y; 
					m_SelectedCP->SetRedo(CurCords);
					m_SelectedCP->m_isMoved = false;
				}
				
				m_SelectedCP->Paste(OldCords.m_x, OldCords.m_y,0,0);
				
				m_SelectedCP->SetUndoToggle(false);
			
			} else {
				OldCords = m_SelectedCP->GetRedo();	
				m_SelectedCP->Paste(OldCords.m_x, OldCords.m_y,0,0);
				m_SelectedCP->SetUndoToggle(true);
			}




	} 
	else  //Move if selected box 3D object
	{
		if (m_FromUndo == ID_FROM_CUT)
		{
	
			m_SelectedBox =  m_Mem.OnPaste(m_Mem.GetUndo().m_x, m_Mem.GetUndo().m_y, m_SceneX, m_SceneY, m_ScreenWidth, m_ScreenHeight, m_CountObject++);
			if (m_SelectedBox) // Add to the 3D object list
			{	
				m_ListBox->Append(m_SelectedBox);
				m_InCopy = false;
			}
			m_FromUndo = ID_FROM_NOTHING;
		} else if (m_SelectedBox)
		{
			Tool::COORDS CurCords, OldCords;

			if (m_SelectedBox->isUndoToggle()) // Toggle UNDO / REDO
			{
				OldCords = m_SelectedBox->GetUndo();
				
				if (m_SelectedBox->m_isMoved) // A new position happens
				{
					CurCords.m_x = m_SelectedBox->GetCoords().m_x;
					CurCords.m_y = m_SelectedBox->GetCoords().m_y; 
					m_SelectedBox->SetRedo(CurCords);
					m_SelectedBox->m_isMoved = false;
				}
				
				m_SelectedBox->Paste(OldCords.m_x, OldCords.m_y,0,0);
				
				m_SelectedBox->SetUndoToggle(false);
			
			} else {
				OldCords = m_SelectedBox->GetRedo();	
				m_SelectedBox->Paste(OldCords.m_x, OldCords.m_y,0,0);
				m_SelectedBox->SetUndoToggle(true);
			}

			m_FromUndo = ID_FROM_NOTHING;
		}
	} 
	
}



// Creates 3D text

void World::OnText(wxString& Str)
{
	Text* txt;
	m_SelectedBox = txt = new Text(m_CountObject++);
	m_ListBox->Append(m_SelectedBox);
	txt->SetText(Str);
	txt->ChangeCoords(0.0f, 0.0f, 1.0f);
	txt->ChangeDimensions(2.0f, 2.0f);
	m_Frame->SetStatusText(_("Change dimension or move"),4);
	m_InAction = true;
}

// Changes color

void World::OnColor(wxColour& Color)
{
	if (m_SelectedBox)
		m_SelectedBox->SetColor(Color);
}

// Changes alpha channel

void World::OnAlpha(int Range)
{
	if (m_SelectedBox)
		m_SelectedBox->SetAlpha(Range);
}

// Refresh combo

void World::OnComboObject(wxComboBox& Combo)
{
	wxString StrObj;

	if (m_ListBox)
	{
		for(ListBox::iterator it = m_ListBox->begin();it != m_ListBox->end();it++)
		{	
			Box *box = *it;
		
			if (box->GetName() == 99999999) // The coordinate system Id
				Combo.Insert(_("Coordinate system::99999999"),0);
			else // The rest of objects
			{
				if (dynamic_cast<Cone*>(box)) StrObj = _("Cone");
				else if (dynamic_cast<Sphere*>(box)) StrObj = _("Sphere");
				else if (dynamic_cast<Cylinder*>(box)) StrObj = _("Cylinder");
				else if (dynamic_cast<Teapot*>(box)) StrObj = _("Teapot");
				else if (dynamic_cast<NURBS*>(box)) StrObj = _("NURBS");
				else if (dynamic_cast<BEZIER*>(box)) StrObj = _("BEZIER");
				else if (dynamic_cast<Text*>(box)) StrObj = _("Text");
				else if (dynamic_cast<Tetrahedron*>(box)) StrObj = _("Tetrahedron");
				else if (dynamic_cast<Octahedron*>(box)) StrObj = _("Octahedron");
				else if (dynamic_cast<Dodecahedron*>(box)) StrObj = _("Dodecahedron");
				else if (dynamic_cast<Icosahedron*>(box)) StrObj = _("Icosahedron");
				else if (dynamic_cast<RombicDodecahedron*>(box)) StrObj = _("RhombicDodecahedron");
				else if (dynamic_cast<Torus*>(box)) StrObj = _("Torus");
				Combo.Insert(StrObj + _("::") +  Tool::Math::ConvertNumber(static_cast<int>(box->GetName())),0);
			}
		}
	}
}

// Combo selection

void World::OnComboSelect(wxString& Select)
{
	const GLuint Name = wxAtoi((Select.SubString(Select.Find(_("::")) + 2, Select.Length())));
	for(ListBox::iterator it = m_ListBox->begin();it != m_ListBox->end();it++) // Iterates all objects
	{	
		Box *box = *it;
		
		if (box->GetName() == Name)
		{
				m_SelectedBox = box;
				m_SceneX = m_SelectedBox->GetCoords().m_x;
				m_SceneY = m_SelectedBox->GetCoords().m_y;
				break;
		}
	}
}

// Handles menu main events

void World::OnMenu(int Code)
{
	Cylinder *cyl;
	Sphere *sph;
	Cone *con;
	Teapot* tp;
	Tetrahedron* th;
	Octahedron* oh;
	Dodecahedron* dh;
	Icosahedron* ih;
	RombicDodecahedron* rdh;
	Torus* tor;

	switch(Code)
	{
		case ID_NURBS: // The user creates a Nurbs
				m_SelectedBox = m_Surface = new NURBS(MAX_CTRL_POINTS,m_CountObject++);
				m_Surface->SetCtrlPoints(new ControlPoint());
				m_ListBox->Append(m_Surface);
				m_Frame->SetStatusText(_("Add control points"),4);
				m_InAction = true;
				break;
		 
		case ID_BEZIER: // The user creates a Nurbs
				m_SelectedBox = m_Surface = new BEZIER(MAX_CTRL_POINTS, m_CountObject++);
				m_Surface->SetCtrlPoints(new ControlPoint());
				m_ListBox->Append(m_Surface);
				m_Frame->SetStatusText(_("Add control points"),4);
				m_InAction = true;
				break;	
		case ID_CYLINDER: // The user creates a cylinder
				m_SelectedBox = cyl = new Cylinder(m_CountObject++);
				m_ListBox->Append(m_SelectedBox);
				cyl->ChangeCoords(m_SceneX, m_SceneY, 1.0f);
				cyl->ChangeDimensions(2.0f,5.0f);
				m_Frame->SetStatusText(_("Change dimensions or move"),4);
				m_InAction = true;
				break;
		case ID_SPHERE: // The user creates a sphere
				m_SelectedBox = sph = new Sphere(m_CountObject++);
				m_ListBox->Append(m_SelectedBox);
				sph->ChangeCoords(m_SceneX, m_SceneY, 1.0f);
				sph->ChangeDimensions(2.0f);
				m_Frame->SetStatusText(_("Change dimensions or move"),4);
				m_InAction = true;
				break;
		case ID_CONE: // The user creates a cone
				m_SelectedBox = con = new Cone(m_CountObject++);
				m_ListBox->Append(m_SelectedBox);
				con->ChangeCoords(m_SceneX, m_SceneY, 1.0f);
				con->ChangeDimensions(2.0f,5.0f);
				m_Frame->SetStatusText(_("Change dimension or move"),4);
				m_InAction = true;
				break;
		case ID_TEAPOT:
				m_SelectedBox = tp = new Teapot(m_CountObject++);
				m_ListBox->Append(m_SelectedBox);
				tp->ChangeCoords(m_SceneX, m_SceneY, 1.0f);
				tp->ChangeDimensions(2.0f);
				m_Frame->SetStatusText(_("Change dimension or move"),4);
				m_InAction = true;
				break;
		case ID_TETRAHEDRON:
				m_SelectedBox = th = new Tetrahedron(m_CountObject++);
				m_ListBox->Append(m_SelectedBox);
				th->ChangeCoords(m_SceneX, m_SceneY, 1.0f);
				th->ChangeDimensions(2.0f);
				m_Frame->SetStatusText(_("Change dimension or move"),4);
				m_InAction = true;
				break;
		case ID_OCTAHEDRON:
				m_SelectedBox = oh = new Octahedron(m_CountObject++);
				m_ListBox->Append(m_SelectedBox);
				oh->ChangeCoords(m_SceneX, m_SceneY, 1.0f);
				oh->ChangeDimensions(2.0f);
				m_Frame->SetStatusText(_("Change dimension or move"),4);
				m_InAction = true;
				break;
		case ID_DODECAHEDRON:
				m_SelectedBox = dh = new Dodecahedron(m_CountObject++);
				m_ListBox->Append(m_SelectedBox);
				dh->ChangeCoords(m_SceneX,m_SceneY, 1.0f);
				dh->ChangeDimensions(2.0f);
				m_Frame->SetStatusText(_("Change dimension or move"),4);
				m_InAction = true;
				break;
		case ID_ICOSAHEDRON:
				m_SelectedBox = ih = new Icosahedron(m_CountObject++);
				m_ListBox->Append(m_SelectedBox);
				ih->ChangeCoords(m_SceneX, m_SceneY, 1.0f);
				ih->ChangeDimensions(2.0f);
				m_Frame->SetStatusText(_("Change dimension or move"),4);
				m_InAction = true;
				break;
		case ID_ROMBIC_DODECAHEDRON:
				m_SelectedBox = rdh = new RombicDodecahedron(m_CountObject++);
				m_ListBox->Append(m_SelectedBox);
				rdh->ChangeCoords(m_SceneX, m_SceneY, 1.0f);
				rdh->ChangeDimensions(2.0f);
				m_Frame->SetStatusText(_("Change dimension or move"),4);
				m_InAction = true;
				break;
		case ID_TORUS:
				m_SelectedBox = tor = new Torus(m_CountObject++);
				m_ListBox->Append(m_SelectedBox);
				tor->ChangeCoords(m_SceneX, m_SceneY, 1.0f);
				tor->ChangeDimensions(2.0f, 2.0f);
				m_Frame->SetStatusText(_("Change dimension or move"),4);
				m_InAction = true;
				break;					
		case wxID_COPY: // The user wants to copy
				OnCopyPaste(ID_COPY);
				m_Frame->SetStatusText(_("Ready to paste"),4);
				m_InAction = true;
				break;
		case wxID_PASTE: // The users does paste
				OnCopyPaste(ID_PASTE);
				break;
		case wxID_CUT: // The user wants to cut
				OnCopyPaste(ID_CUT);
				m_Frame->SetStatusText(_("Ready to paste"),4);
				m_InAction = true;
				break;
		case wxID_DELETE: // The users deletes an object
				DeleteSelected();
				m_Frame->SetStatusText(_("Object deleted"),4);
				m_InAction = true;
				break;
		case wxID_TRANSFORM: // The users transforms a quadric
				if ((m_SelectedBox) && (dynamic_cast<Quadric*>(m_SelectedBox)))
				{
					m_InModelMove = true;
					m_InAction = false;	
					m_Frame->SetStatusText(_("Transforming. Press right button to finish"),4);
				}
				break;
		case ID_UNDO:
				OnDoUndoRedo();
				m_Frame->SetStatusText(_("Just undone!"),4);
				m_InAction = true;
				break;
		case ID_REDO:
				OnDoUndoRedo();
				m_Frame->SetStatusText(_("Just redone!"),4);
				m_InAction = true;
				break;
	}

}

// XML file save

bool World::OnXMLFileSave(wxString& XMLContent)
{
	
	m_Frame->SetStatusText(_("Saving file..."),4);
	try{ // Steps out the coordiante system
		for(ListBox::iterator it = ++m_ListBox->begin();it != m_ListBox->end();it++)
		{
			(*it)->Save(XMLContent); // Call base class 
		}
	} catch(std::exception &ex)
	{
		wxMessageBox(_("Error saving XML ISSI file. Error description:") + _(ex.what()), _("XML save error"), wxOK | wxICON_ERROR);
		m_Frame->SetStatusText(_("Error when saving"),4);
		return false;
	}
	
	m_Frame->SetStatusText(_("File saved OK"),4);
	
	return true;
}

// XML file open

bool World::OnXMLFileOpen(wxString& XMLContent)
{
	
	Destroy(); // Reinitiate world
	Create();   	
	
	m_Frame->SetStatusText(_("Opening file..."),4);
	
	try{
		xml_document<> doc;    // character type defaults to char
		wxCharBuffer buffer = XMLContent.ToUTF8();
		doc.parse<0>(buffer.data());    // 0 means default parse flags
		xml_node<>* root_node;
		root_node = doc.first_node();
		
		// For each issiform

		for(xml_node<>* node = root_node->first_node("issiform"); node; node = node->next_sibling())
		{
		
			Box* Box = NULL;
			if (!strcmp(node->first_attribute()->value(),"Cylinder"))
			{
				Box = new Cylinder(m_CountObject++);
				m_ListBox->Append(Box);
			} else if (!strcmp(node->first_attribute()->value(),"Sphere"))
			{
				Box = new Sphere(m_CountObject++);
				m_ListBox->Append(Box);
			} else if (!strcmp(node->first_attribute()->value(),"Cone"))
			{
				Box = new Cone(m_CountObject++);
				m_ListBox->Append(Box);
			}
			else if (!strcmp(node->first_attribute()->value(),"Teapot"))
			{
				Box = new Teapot(m_CountObject++);
				m_ListBox->Append(Box);
			} 
			else if (!strcmp(node->first_attribute()->value(),"Text"))
			{
				Box = new Text(m_CountObject++);
				m_ListBox->Append(Box);
			}
			else if (!strcmp(node->first_attribute()->value(),"Tetrahedron"))
			{
				Box = new Tetrahedron(m_CountObject++);
				m_ListBox->Append(Box);
			} 
			else if (!strcmp(node->first_attribute()->value(),"Octahedron"))
			{
				Box = new Octahedron(m_CountObject++);
				m_ListBox->Append(Box);
			} 
			else if (!strcmp(node->first_attribute()->value(),"Dodecahedron"))
			{
				Box = new Dodecahedron(m_CountObject++);
				m_ListBox->Append(Box);
			} 
			else if (!strcmp(node->first_attribute()->value(),"Icosahedron"))
			{
				Box = new Icosahedron(m_CountObject++);
				m_ListBox->Append(Box);
			} 
			else if (!strcmp(node->first_attribute()->value(),"RombicDodecahedron"))
			{
				Box = new RombicDodecahedron(m_CountObject++);
				m_ListBox->Append(Box);
			} 
			else if (!strcmp(node->first_attribute()->value(),"Torus"))
			{
				Box = new Torus(m_CountObject++);
				m_ListBox->Append(Box);
			} 
			else if (!strcmp(node->first_attribute()->value(),"Nurbs"))
			{
				
				Box = new NURBS(MAX_CTRL_POINTS, m_CountObject++);
				Surface* Surf = dynamic_cast<Surface*>(Box);
				Surf->SetCtrlPoints(new ControlPoint());
				m_ListBox->Append(Surf);
				m_CountObject = Surf->CtrlCPLoad(node->first_node()->next_sibling());
			}  else if (!strcmp(node->first_attribute()->value(),"Bezier"))
			{
				Box = new BEZIER(MAX_CTRL_POINTS, m_CountObject++);
				Surface* Surf = dynamic_cast<Surface*>(Box);
				Surf->SetCtrlPoints(new ControlPoint());
				m_ListBox->Append(Box);
				m_CountObject = Surf->CtrlCPLoad(node->first_node());
			}
			if (!strcmp(node->first_node()->name(),"issiobjectcolor"))
				Box->Load(node->first_node()->first_node());				
			else Box->Load(node->first_node());				
		}
	} catch (std::exception& ex)
	{
		wxMessageBox(_("Error opening XML ISSI file. Error description:") + _(ex.what()), _("XML open error"), wxOK | wxICON_ERROR);
		m_Frame->SetStatusText(_("Error when opening"),4);
		return false;
	}
	m_Frame->SetStatusText(_("File open OK"),4);
	return true;
}

// Destroy all 3D object in the World object

void World::Destroy()
{
	

	for(ListBox::iterator it = m_ListBox->begin();it != m_ListBox->end();it++)
	{
		Box *box = *it;
		delete box;
	}
		
	delete m_ListBox;
}

// Destructor

World::~World()
{
	Destroy();
	
}
