/*****************************************************************************
   ISSIGraph: Experimental Surface Editor Prototype						 
   Copyright (C) 2013 by Carlos Jimnez de Parga								 
   
   Trabajo para el:
   MASTER EN INVESTIGACIN EN INGENIERA DE SOFTWARE Y SISTEMAS INFORMTICOS
			REPRESENTACIN GRFICA DE SUPERFICIES IMPLCITAS			
	DEPARTAMENTO DE INGENIERA DE SOFTWARE Y SISTEMAS INFORMTICOS 
							- UNED -

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
   **************************************************************************/

/* 
	FILE: World.h
	PURPOSE: World and render bussines logic (header)
	Author: Carlos Jimnez de Parga (c)
	License: GPL v3.0
*/

#ifndef _WORLD_H_
#define _WORLD_H_
         
#include "box.h"
#include "surface.h"
#include "quadric.h"
#include "textout.h"
#include "windows.h"
#include "definitions.h"
#include "wx/wx.h"
#include "wx/dialog.h"

// XML wrapper

using namespace rapidxml;

// ----------------------------------------------------------------------------
// Constants
// ----------------------------------------------------------------------------

const float ANGLE_SPIN = 5.0f;
const float ANGLE_SCENE = 2.0f;
const int   MAX_CTRL_POINTS = 64;

// ----------------------------------------------------------------------------
// Axis class
// ----------------------------------------------------------------------------

class Axis : public VertexBox
{
public:
	Axis(GLuint name, float ptx, float pty, float ptz, float width, float height);
	void Draw();
	void Spin(float xSpin, float ySpin);
	float GetAngleX();
	float GetAngleY();
	void SetAngleX(float Xangle);
	void SetAngleY(float Yangle);
private:
};


// Forward class declaration
class World;

// ----------------------------------------------------------------------------
// Memory class
// ----------------------------------------------------------------------------

class Memory
{
public:
	Memory();
	~Memory();
	void OnCopy(Box* Selected, GLuint Count);
	Box* OnPaste(int x, int y, float SceneX, float SceneY, int ScrW, int ScrH, GLuint Count);
	void SetMouse(int x, int y);
	Tool::COORDS GetMouse();
	Tool::COORDS GetUndo();
	void SetUndo(Tool::COORDS Cords);
private:
	Box* switchClass(GLuint Count);
	Box* m_Mem, *m_Cut;
	int m_MouseX, m_MouseY;
	int m_UndoX, m_UndoY;
};

// Declare list of Box objects

WX_DECLARE_LIST(Box, ListBox);

// ----------------------------------------------------------------------------
// World class: Contains the scene
// ----------------------------------------------------------------------------

class World 
{
public:
	World(wxGLCanvas* Canvas, wxFrame* Frame);
	~World();
	void Create();
	void Build();
	void Render(bool Lactive);
	void OnMouseRDown(int x, int y, Tool::Idents* Idents, int ScreenWidth, int ScreenHeight);
	void OnMouseRDown(int Id);
	void OnMouseLDown(Tool::Idents* Idents);
	void OnMouseMove(int x, int y, int ScreenWidth, int ScreenHeight);
	void OnMouseLUp(int x, int y);
	void OnMouseRUp();
	void OnMouseWheel(int Zooming);
	bool OnKeyDown(int KeyCode);
	void OnChangeSize();
	void OnDoUndoRedo();
	void OnText(wxString& Str);
	void OnColor(wxColour& Color);
	void OnAlpha(int Range);
	void OnComboObject(wxComboBox& Combo);
	void OnComboSelect(wxString& Select);
	void OnSetControlOrder(int OrderId);
	void DrawSurface();
	bool Detect(int x, int y);
	void OnMenu(int Code);
	void OnSaving(bool Saving);
	bool GetDbBox();
	bool GetInModeling();
	bool GetInModelMove();
	void DeleteSelected();
	bool OnXMLFileSave(wxString& XMLContent);
	bool OnXMLFileOpen(wxString& XMLContent);
	void RotateScene();
	void Destroy();
private:
	
	void ChangeTwoDim(int x, int y, TwoDimension* Cyl);
	void ChangeOneDim(int x, OneDimension* Sph);
	void OnCopyPaste(int Id);
	void OnSpin(float XSpin,float YSpin);

	bool m_ButtonRDown, m_ButtonLDown, m_ZoomPlus;
	bool m_InModeling, m_InModelMove, m_InAction;
	bool m_Saving, m_InCopy, m_ComeFromLocal, m_ComeFromClick;
	int m_MouseX, m_MouseY, m_RotateX, m_RotateY, m_ScreenWidth, m_ScreenHeight;
	GLuint m_CountObject;
	
	float m_SceneX, m_SceneY, m_DepthZ;
	enum State {ID_FROM_NOTHING, ID_FROM_COPY, ID_FROM_CUT};
	State m_FromUndo;
	wxFrame* m_Frame;
	wxGLCanvas* m_Canvas; // Owner canvas object
	Axis *m_CordSys; // The coordinate system
	CtrlPoint* m_SelectedCP; // Selected control point
	Box* m_SelectedBox; // Selected 3D object
	Memory m_Mem; // Copy/cut/paste object
	ListBox* m_ListBox; // List of 3D objects
	Surface* m_Surface; // Surface
	Tool::COORDS m_UndoCords; // Remembers old undo left clickdown
};


#endif // _WORLD_H_